"""Well grid information."""
import collections
from dataclasses import dataclass
from typing import Dict, List

from opentrons_shared_data.labware.constants import WELL_NAME_PATTERN


@dataclass(frozen=True)
class WellGrid:
    """Well grid information.

    Attributes:
        columns_by_name: An ordered mapping of well names, keyed by column names.
        rows_by_name: An ordered mapping of well names, keyed by row names.
    """

    rows_by_name: Dict[str, List[str]]
    columns_by_name: Dict[str, List[str]]


# NOTE(mc, 2022-11-11): to preserve historical behavior,
# this function organizes rows and columns using the well name,
# rather than the well's position in the labware definition's `ordering` field.
# Logic originally added in https://github.com/Opentrons/opentrons/pull/2418
def create(columns: List[List[str]]) -> WellGrid:
    columns_by_name = collections.defaultdict(list)
    rows_by_name = collections.defaultdict(list)

    for column in columns:
        for well_name in column:
            well_name_match = WELL_NAME_PATTERN.match(well_name)

            assert (
                well_name_match is not None
            ), "Well name did not match required pattern; please check labware definition."

            row_name, column_name = well_name_match.group(1, 2)
            rows_by_name[row_name].append(well_name)
            columns_by_name[column_name].append(well_name)

    return WellGrid(
        columns_by_name=columns_by_name,
        rows_by_name=rows_by_name,
    )



"""Base command data model and type definitions."""


from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING, Generic, Optional, TypeVar, Tuple

from pydantic import BaseModel, Field
from pydantic.generics import GenericModel

from opentrons.hardware_control import HardwareControlAPI

from ..errors import ErrorOccurrence

# Work around type-only circular dependencies.
if TYPE_CHECKING:
    from .. import execution
    from ..state import StateView


CommandParamsT = TypeVar("CommandParamsT", bound=BaseModel)

CommandResultT = TypeVar("CommandResultT", bound=BaseModel)

CommandPrivateResultT = TypeVar("CommandPrivateResultT")


class CommandStatus(str, Enum):
    """Command execution status."""

    QUEUED = "queued"
    RUNNING = "running"
    SUCCEEDED = "succeeded"
    FAILED = "failed"


class CommandIntent(str, Enum):
    """Run intent for a given command.

    Props:
        PROTOCOL: the command is part of the protocol run itself.
        SETUP: the command is part of the setup phase of a run.
    """

    PROTOCOL = "protocol"
    SETUP = "setup"


class BaseCommandCreate(GenericModel, Generic[CommandParamsT]):
    """Base class for command creation requests.

    You shouldn't use this class directly; instead, use or define
    your own subclass per specific command type.
    """

    commandType: str = Field(
        ...,
        description=(
            "Specific command type that determines data requirements and "
            "execution behavior"
        ),
    )
    params: CommandParamsT = Field(..., description="Command execution data payload")
    intent: Optional[CommandIntent] = Field(
        None,
        description=(
            "The reason the command was added. If not specified or `protocol`,"
            " the command will be treated as part of the protocol run itself,"
            " and added to the end of the existing command queue."
            "\n\n"
            "If `setup`, the command will be treated as part of run setup."
            " A setup command may only be enqueued if the run has not started."
            "\n\n"
            "Use setup commands for activities like pre-run calibration checks"
            " and module setup, like pre-heating."
        ),
    )
    key: Optional[str] = Field(
        None,
        description=(
            "A key value, unique in this run, that can be used to track"
            " the same logical command across multiple runs of the same protocol."
            " If a value is not provided, one will be generated."
        ),
    )


class BaseCommand(GenericModel, Generic[CommandParamsT, CommandResultT]):
    """Base command model.

    You shouldn't use this class directly; instead, use or define
    your own subclass per specific command type.
    """

    id: str = Field(
        ...,
        description="Unique identifier of this particular command instance",
    )
    createdAt: datetime = Field(..., description="Command creation timestamp")
    commandType: str = Field(
        ...,
        description=(
            "Specific command type that determines data requirements and "
            "execution behavior"
        ),
    )
    key: str = Field(
        ...,
        description=(
            "An identifier representing this command as a step in a protocol."
            " A command's `key` will be unique within a given run, but stable"
            " across all runs that perform the same exact procedure. Thus,"
            " `key` be used to compare/match commands across multiple runs"
            " of the same protocol."
        ),
    )
    status: CommandStatus = Field(..., description="Command execution status")
    params: CommandParamsT = Field(..., description="Command execution data payload")
    result: Optional[CommandResultT] = Field(
        None,
        description="Command execution result data, if succeeded",
    )
    error: Optional[ErrorOccurrence] = Field(
        None,
        description="Reference to error occurrence, if execution failed",
    )
    startedAt: Optional[datetime] = Field(
        None,
        description="Command execution start timestamp, if started",
    )
    completedAt: Optional[datetime] = Field(
        None,
        description="Command execution completed timestamp, if completed",
    )
    intent: Optional[CommandIntent] = Field(
        None,
        description=(
            "The reason the command was added to the run."
            " If not specified or `protocol`, it is part of the protocol itself."
            " If `setup`, it was added as part of setup; for example,"
            " a command that is part of a calibration procedure."
        ),
    )


class AbstractCommandImpl(
    ABC,
    Generic[CommandParamsT, CommandResultT],
):
    """Abstract command creation and execution implementation.

    A given command request should map to a specific command implementation,
    which defines how to:

    - Create a command resource from the request model
    - Execute the command, mapping data from execution into the result model

    This class should be used as the base class for new commands by default. You should only
    use AbstractCommandWithPrivateResultImpl if you actually need private results to send to
    the rest of the engine wihtout being published outside of it.
    """

    def __init__(
        self,
        state_view: StateView,
        hardware_api: HardwareControlAPI,
        equipment: execution.EquipmentHandler,
        movement: execution.MovementHandler,
        gantry_mover: execution.GantryMover,
        labware_movement: execution.LabwareMovementHandler,
        pipetting: execution.PipettingHandler,
        tip_handler: execution.TipHandler,
        run_control: execution.RunControlHandler,
        rail_lights: execution.RailLightsHandler,
        status_bar: execution.StatusBarHandler,
    ) -> None:
        """Initialize the command implementation with execution handlers."""
        pass

    @abstractmethod
    async def execute(self, params: CommandParamsT) -> CommandResultT:
        """Execute the command, mapping data from execution into a response model."""
        ...


class AbstractCommandWithPrivateResultImpl(
    ABC,
    Generic[CommandParamsT, CommandResultT, CommandPrivateResultT],
):
    """Abstract command creation and execution implementation if the command has private results.

    A given command request should map to a specific command implementation,
    which defines how to:

    - Create a command resource from the request model
    - Execute the command, mapping data from execution into the result model

    This class should be used instead of AbstractCommandImpl as a base class if your command needs
    to send data to result handlers that should not be published outside of the engine.

    Note that this class needs an extra type-parameter for the private result.
    """

    def __init__(
        self,
        state_view: StateView,
        hardware_api: HardwareControlAPI,
        equipment: execution.EquipmentHandler,
        movement: execution.MovementHandler,
        gantry_mover: execution.GantryMover,
        labware_movement: execution.LabwareMovementHandler,
        pipetting: execution.PipettingHandler,
        tip_handler: execution.TipHandler,
        run_control: execution.RunControlHandler,
        rail_lights: execution.RailLightsHandler,
        status_bar: execution.StatusBarHandler,
    ) -> None:
        """Initialize the command implementation with execution handlers."""
        pass

    @abstractmethod
    async def execute(
        self, params: CommandParamsT
    ) -> Tuple[CommandResultT, CommandPrivateResultT]:
        """Execute the command, mapping data from execution into a response model."""
        ...



import logging
from typing import Dict, List, Optional, Set, Union, cast, Tuple

from opentrons_shared_data.deck.dev_types import DeckDefinitionV4, SlotDefV3
from opentrons_shared_data.labware.dev_types import LabwareDefinition
from opentrons_shared_data.pipette.dev_types import PipetteNameType
from opentrons_shared_data.robot.dev_types import RobotType

from opentrons.types import DeckSlotName, StagingSlotName, Location, Mount, Point
from opentrons.util.broker import Broker
from opentrons.hardware_control import SyncHardwareAPI
from opentrons.hardware_control.modules import AbstractModule, ModuleModel, ModuleType
from opentrons.hardware_control.types import DoorState, PauseType
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.api_support.util import AxisMaxSpeeds, APIVersionError
from opentrons.protocols import labware as labware_definition

from ...labware import Labware
from ..._liquid import Liquid
from ..._types import OffDeckType
from ..._trash_bin import TrashBin
from ..._waste_chute import WasteChute
from ..protocol import AbstractProtocol
from ..labware import LabwareLoadParams

from . import legacy_module_core, module_geometry
from .deck import Deck
from .legacy_instrument_core import LegacyInstrumentCore
from .labware_offset_provider import AbstractLabwareOffsetProvider
from .legacy_labware_core import LegacyLabwareCore
from .load_info import LoadInfo, InstrumentLoadInfo, LabwareLoadInfo, ModuleLoadInfo

logger = logging.getLogger(__name__)


class LegacyProtocolCore(
    AbstractProtocol[
        LegacyInstrumentCore,
        LegacyLabwareCore,
        legacy_module_core.LegacyModuleCore,
    ]
):
    def __init__(
        self,
        sync_hardware: SyncHardwareAPI,
        api_version: APIVersion,
        labware_offset_provider: AbstractLabwareOffsetProvider,
        deck_layout: Deck,
        equipment_broker: Optional[Broker[LoadInfo]] = None,
        bundled_labware: Optional[Dict[str, LabwareDefinition]] = None,
        extra_labware: Optional[Dict[str, LabwareDefinition]] = None,
    ) -> None:
        """Build a :py:class:`.LegacyProtocolCore`.

        :param api_version: The API version to use. If this is ``None``, uses
                            the max supported version.
        :param hardware: The hardware control API to use.
        :param bundled_labware: A dict mapping labware URIs to definitions.
                                This is used when executing bundled protocols,
                                and if specified will be the only allowed
                                source for labware definitions, excluding the
                                built in definitions and anything in
                                ``extra_labware``.
        :param bundled_data: A dict mapping filenames to the contents of data
                             files. Can be used by the protocol, since it is
                             exposed as
                             :py:attr:`.ProtocolContext.bundled_data`
        :param extra_labware: A dict mapping labware URIs to definitions. These
                              URIs are searched during :py:meth:`.load_labware`
                              in addition to the system definitions (if
                              ``bundled_labware`` was not specified). Used to
                              provide custom labware definitions.
        """
        self._sync_hardware = sync_hardware
        self._api_version = api_version
        self._labware_offset_provider = labware_offset_provider
        self._deck_layout = deck_layout
        self._equipment_broker = equipment_broker or Broker()

        self._instruments: Dict[Mount, Optional[LegacyInstrumentCore]] = {
            mount: None for mount in Mount.ot2_mounts()  # Legacy core works only on OT2
        }
        self._bundled_labware = bundled_labware
        self._extra_labware = extra_labware or {}
        self._default_max_speeds = AxisMaxSpeeds(robot_type=self.robot_type)
        self._last_location: Optional[Location] = None
        self._last_mount: Optional[Mount] = None
        self._loaded_modules: Set["AbstractModule"] = set()
        self._module_cores: List[legacy_module_core.LegacyModuleCore] = []
        self._labware_cores: List[LegacyLabwareCore] = [self.fixed_trash]
        self._disposal_locations: List[Union[Labware, TrashBin, WasteChute]] = []

    @property
    def api_version(self) -> APIVersion:
        """Get the API version the protocol is adhering to."""
        return self._api_version

    @property
    def robot_type(self) -> RobotType:
        return "OT-2 Standard"

    @property
    def equipment_broker(self) -> Broker[LoadInfo]:
        """A message broker to to publish equipment load events.

        Subscribers to this broker will be notified with information about every
        successful labware load, instrument load, or module load.

        Only this interface is allowed to publish to this broker.
        Calling code may only subscribe or unsubscribe.
        """
        return self._equipment_broker

    def get_bundled_labware(self) -> Optional[Dict[str, LabwareDefinition]]:
        """Bundled labware definition."""
        # TODO AL 20201110 - This should be removed along with the bundling
        #  feature as we move to HTTP based protocol execution.
        return self._bundled_labware

    def get_extra_labware(self) -> Optional[Dict[str, LabwareDefinition]]:
        """Extra labware definitions."""
        return self._extra_labware

    def get_max_speeds(self) -> AxisMaxSpeeds:
        """Get the maximum axis speeds."""
        return self._default_max_speeds

    def get_hardware(self) -> SyncHardwareAPI:
        """Access to the synchronous hardware API."""
        return self._sync_hardware

    def is_simulating(self) -> bool:
        """Returns true if hardware is being simulated."""
        return self._sync_hardware.is_simulator  # type: ignore[no-any-return]

    def append_disposal_location(
        self,
        disposal_location: Union[Labware, TrashBin, WasteChute],
    ) -> None:
        if isinstance(disposal_location, (TrashBin, WasteChute)):
            raise APIVersionError(
                "Trash Bin and Waste Chute Disposal locations are not supported in this API Version."
            )
        self._disposal_locations.append(disposal_location)

    def add_disposal_location_to_engine(
        self, disposal_location: Union[TrashBin, WasteChute]
    ) -> None:
        assert False, "add_disposal_location_to_engine only supported on engine core"

    def add_labware_definition(
        self,
        definition: LabwareDefinition,
    ) -> LabwareLoadParams:
        """Add a labware definition to the set of loadable definitions."""
        load_params = LabwareLoadParams(
            namespace=definition["namespace"],
            load_name=definition["parameters"]["loadName"],
            version=definition["version"],
        )
        self._extra_labware = self._extra_labware.copy()
        self._extra_labware[load_params.as_uri()] = definition
        return load_params

    def load_labware(
        self,
        load_name: str,
        location: Union[
            DeckSlotName,
            LegacyLabwareCore,
            legacy_module_core.LegacyModuleCore,
            StagingSlotName,
            OffDeckType,
        ],
        label: Optional[str],
        namespace: Optional[str],
        version: Optional[int],
    ) -> LegacyLabwareCore:
        """Load a labware using its identifying parameters."""
        if isinstance(location, OffDeckType):
            raise APIVersionError(
                "Loading a labware off deck is only supported with apiLevel 2.15 and newer."
            )
        elif isinstance(location, LegacyLabwareCore):
            raise APIVersionError(
                "Loading a labware onto another labware or adapter is only supported with api version 2.15 and above"
            )
        elif isinstance(location, StagingSlotName):
            raise APIVersionError("Using a staging deck slot requires apiLevel 2.16.")

        deck_slot = (
            location if isinstance(location, DeckSlotName) else location.get_deck_slot()
        )

        parent = (
            self.get_deck().position_for(location.value)
            if isinstance(location, DeckSlotName)
            else location.geometry.location
        )

        labware_def = labware_definition.get_labware_definition(
            load_name,
            namespace,
            version,
            bundled_defs=self._bundled_labware,
            extra_defs=self._extra_labware,
        )
        labware_core = LegacyLabwareCore(
            definition=labware_def,
            parent=parent,
            label=label,
        )
        labware_load_params = labware_core.get_load_params()
        labware_offset = self._labware_offset_provider.find(
            load_params=labware_load_params,
            deck_slot=deck_slot,
            requested_module_model=(
                location.get_requested_model()
                if isinstance(location, legacy_module_core.LegacyModuleCore)
                else None
            ),
        )
        labware_core.set_calibration(labware_offset.delta)

        self._labware_cores.append(labware_core)
        if isinstance(location, DeckSlotName):
            # This assignment will raise if the new item conflicts with something else
            # on the deck--for example, if something tall is placed next to a
            # Heater-Shaker.
            #
            # It's a latent bug that we only do this conflict checking when loading
            # directly into a deck slot. We should also do conflict checking when
            # labware is loaded atop a module, because that affects the module's
            # maximum height.
            #
            # In practice, I don't think this matters now (2023-02-22) because of the
            # exact conflict checks that we perform. Wherever we have a constraint on
            # maximum height, we also happen to have a constraint disallowing modules
            # in the first place.
            self._deck_layout[location.value] = labware_core

        self._equipment_broker.publish(
            LabwareLoadInfo(
                labware_definition=labware_core.get_definition(),
                labware_namespace=labware_load_params.namespace,
                labware_load_name=labware_load_params.load_name,
                labware_version=labware_load_params.version,
                deck_slot=deck_slot,
                on_module=isinstance(location, legacy_module_core.LegacyModuleCore),
                offset_id=labware_offset.offset_id,
                labware_display_name=labware_core.get_user_display_name(),
            )
        )

        return labware_core

    def load_adapter(
        self,
        load_name: str,
        location: Union[
            DeckSlotName,
            StagingSlotName,
            legacy_module_core.LegacyModuleCore,
            OffDeckType,
        ],
        namespace: Optional[str],
        version: Optional[int],
    ) -> LegacyLabwareCore:
        """Load an adapter using its identifying parameters"""
        raise APIVersionError("Loading adapter is not supported in this API version.")

    # TODO (spp, 2022-12-14): https://opentrons.atlassian.net/browse/RLAB-237
    def move_labware(
        self,
        labware_core: LegacyLabwareCore,
        new_location: Union[
            DeckSlotName,
            StagingSlotName,
            LegacyLabwareCore,
            legacy_module_core.LegacyModuleCore,
            OffDeckType,
            WasteChute,
        ],
        use_gripper: bool,
        pause_for_manual_move: bool,
        pick_up_offset: Optional[Tuple[float, float, float]],
        drop_offset: Optional[Tuple[float, float, float]],
    ) -> None:
        """Move labware to new location."""
        raise APIVersionError("Labware movement is not supported in this API version")

    def load_module(
        self,
        model: ModuleModel,
        deck_slot: Optional[DeckSlotName],
        configuration: Optional[str],
    ) -> legacy_module_core.LegacyModuleCore:
        """Load a module."""
        resolved_type = ModuleType.from_model(model)
        resolved_location = self._deck_layout.resolve_module_location(
            resolved_type, (None if deck_slot is None else deck_slot.id)
        )

        selected_hardware = None
        selected_definition = None

        for module_hardware in self._sync_hardware.attached_modules:
            if module_hardware not in self._loaded_modules:
                definition = module_geometry.load_definition(module_hardware.model())
                compatible = module_geometry.models_compatible(model, definition)

                if compatible:
                    self._loaded_modules.add(module_hardware)
                    selected_hardware = module_hardware
                    selected_definition = definition
                    break

        if selected_hardware is None and self.is_simulating():
            selected_hardware = self._sync_hardware.create_simulating_module(model)
            selected_definition = module_geometry.load_definition(model)

        if selected_hardware is None or selected_definition is None:
            raise RuntimeError(f"Could not find specified module: {model.value}")

        # Load geometry to match the hardware module that we found connected.
        geometry = module_geometry.create_geometry(
            definition=selected_definition,
            parent=self._deck_layout.position_for(resolved_location),
            configuration=configuration,
        )

        module_core = legacy_module_core.create_module_core(
            module_hardware_api=selected_hardware,
            requested_model=model,
            geometry=geometry,
            protocol_core=self,
        )

        self._deck_layout[resolved_location] = geometry
        self._module_cores.append(module_core)

        self.equipment_broker.publish(
            ModuleLoadInfo(
                requested_model=model,
                loaded_model=module_core.get_model(),
                module_serial=module_core.get_serial_number(),
                deck_slot=module_core.get_deck_slot(),
                configuration=configuration,
            )
        )

        return module_core

    def load_instrument(
        self, instrument_name: PipetteNameType, mount: Mount
    ) -> LegacyInstrumentCore:
        """Load an instrument."""
        attached = {
            att_mount: instr.get("name", None)
            for att_mount, instr in self._sync_hardware.attached_instruments.items()
            if instr
        }
        attached[mount] = instrument_name.value
        self._sync_hardware.cache_instruments(attached)
        # If the cache call didn’t raise, the instrument is attached
        new_instr = LegacyInstrumentCore(
            api_version=self._api_version,
            protocol_interface=self,
            mount=mount,
            instrument_name=instrument_name.value,
            default_speed=400.0,  # TODO(mm, 2022-11-10): Deduplicate with engine core.
        )
        self._instruments[mount] = new_instr
        logger.info("Instrument {} loaded".format(new_instr))

        pipette_dict = self._sync_hardware.get_attached_instrument(mount)
        self._equipment_broker.publish(
            InstrumentLoadInfo(
                instrument_load_name=instrument_name.value,
                mount=mount,
                pipette_dict=pipette_dict,
            )
        )

        return new_instr

    def get_loaded_instruments(
        self,
    ) -> Dict[Mount, Optional[LegacyInstrumentCore]]:
        """Get a mapping of mount to instrument."""
        return self._instruments

    def get_disposal_locations(self) -> List[Union[Labware, TrashBin, WasteChute]]:
        """Get valid disposal locations."""
        return self._disposal_locations

    def pause(self, msg: Optional[str]) -> None:
        """Pause the protocol."""
        self._sync_hardware.pause(PauseType.PAUSE)

    def resume(self) -> None:
        """Result the protocol."""
        self._sync_hardware.resume(PauseType.PAUSE)

    def comment(self, msg: str) -> None:
        """Add comment to run log."""
        pass

    def delay(self, seconds: float, msg: Optional[str]) -> None:
        """Delay execution for x seconds."""
        self._sync_hardware.delay(seconds)

    def home(self) -> None:
        """Home the robot."""
        self.set_last_location(None)
        self._sync_hardware.home()

    def get_deck(self) -> Deck:
        """Get the deck layout."""
        return self._deck_layout

    @property
    def fixed_trash(self) -> LegacyLabwareCore:
        """The trash fixed to slot 12 of the robot deck."""
        trash = self._deck_layout["12"]

        if isinstance(trash, LegacyLabwareCore):
            return trash
        if isinstance(trash, Labware):
            return cast(LegacyLabwareCore, trash._core)

        raise RuntimeError("Robot must have a trash container in 12")

    def set_rail_lights(self, on: bool) -> None:
        """Set the rail light state."""
        self._sync_hardware.set_lights(rails=on)

    def get_rail_lights_on(self) -> bool:
        """Get the rail light state."""
        return self._sync_hardware.get_lights()["rails"]  # type: ignore[no-any-return]

    def door_closed(self) -> bool:
        """Check if door is closed."""
        return DoorState.CLOSED == self._sync_hardware.door_state  # type: ignore[no-any-return]

    def get_last_location(
        self,
        mount: Optional[Mount] = None,
    ) -> Optional[Location]:
        """Get the most recent moved to location."""
        if mount is None or mount == self._last_mount:
            return self._last_location

        return None

    def set_last_location(
        self,
        location: Optional[Location],
        mount: Optional[Mount] = None,
    ) -> None:
        """Set the most recent moved to location."""
        self._last_location = location
        self._last_mount = mount

    def get_module_cores(self) -> List[legacy_module_core.LegacyModuleCore]:
        """Get loaded module cores."""
        return self._module_cores

    def get_labware_cores(self) -> List[LegacyLabwareCore]:
        """Get all loaded labware cores."""
        return self._labware_cores

    def get_labware_on_module(
        self, module_core: legacy_module_core.LegacyModuleCore
    ) -> Optional[LegacyLabwareCore]:
        """Get the item on top of a given module, if any."""
        labware = module_core.geometry.labware
        return cast(LegacyLabwareCore, labware._core) if labware is not None else None

    def get_labware_on_labware(
        self, labware_core: LegacyLabwareCore
    ) -> Optional[LegacyLabwareCore]:
        assert False, "get_labware_on_labware only supported on engine core"

    def get_deck_definition(self) -> DeckDefinitionV4:
        """Get the geometry definition of the robot's deck."""
        assert False, "get_deck_definition only supported on engine core"

    def get_slot_definition(
        self, slot: Union[DeckSlotName, StagingSlotName]
    ) -> SlotDefV3:
        """Get the slot definition from the robot's deck."""
        assert False, "get_slot_definition only supported on engine core"

    def get_slot_definitions(self) -> Dict[str, SlotDefV3]:
        """Get all standard slot definitions available in the deck definition."""
        assert False, "get_slot_definitions only supported on engine core"

    def get_staging_slot_definitions(self) -> Dict[str, SlotDefV3]:
        """Get all staging slot definitions available in the deck definition."""
        assert False, "get_staging_slot_definitions only supported on engine core"

    def get_slot_item(
        self, slot_name: Union[DeckSlotName, StagingSlotName]
    ) -> Union[LegacyLabwareCore, legacy_module_core.LegacyModuleCore, None]:
        """Get the contents of a given slot, if any."""
        assert False, "get_slot_item only supported on engine core"

    def get_slot_center(self, slot_name: Union[DeckSlotName, StagingSlotName]) -> Point:
        """Get the absolute coordinate of a slot's center."""
        assert False, "get_slot_center only supported on engine core."

    def get_highest_z(self) -> float:
        """Get the highest Z point of all deck items."""
        assert False, "get_highest_z only supported on engine core"

    def define_liquid(
        self, name: str, description: Optional[str], display_color: Optional[str]
    ) -> Liquid:
        """Define a liquid to load into a well."""
        assert False, "define_liquid only supported on engine core"

    def get_labware_location(
        self, labware_core: LegacyLabwareCore
    ) -> Union[
        str, LegacyLabwareCore, legacy_module_core.LegacyModuleCore, OffDeckType
    ]:
        """Get labware parent location."""
        assert False, "get_labware_location only supported on engine core"



"""Common configuration command base models."""

from pydantic import BaseModel, Field
from typing import Optional
from dataclasses import dataclass
from opentrons.hardware_control.nozzle_manager import (
    NozzleMap,
)
from ..resources import pipette_data_provider


@dataclass
class PipetteConfigUpdateResultMixin:
    """A mixin-suitable model for adding pipette config to results."""

    pipette_id: str
    serial_number: str
    config: pipette_data_provider.LoadedStaticPipetteData


class PipetteNozzleLayoutResultMixin(BaseModel):
    """A nozzle layout result for updating the pipette state."""

    pipette_id: str
    nozzle_map: Optional[NozzleMap] = Field(
        default=None,
        description="A dataclass object holding information about the current nozzle configuration.",
    )



"""
- This driver is responsible for providing an interface for the mag deck
- The driver is the only system component that knows about the mag-deck's
  GCODES or how the mag-deck communicates

- The driver is NOT responsible for interpreting deck states in any way
  or knowing anything about what the device is being used for
"""

from __future__ import annotations

import asyncio
import logging
from typing import Dict, Optional
from enum import Enum

from opentrons.drivers import utils
from opentrons.drivers.command_builder import CommandBuilder
from opentrons.drivers.asyncio.communication import SerialConnection
from .abstract import AbstractMagDeckDriver

log = logging.getLogger(__name__)

DEFAULT_MAG_DECK_TIMEOUT = 10  # Quite large to account for probe time

DEFAULT_COMMAND_RETRIES = 3


class GCODE(str, Enum):
    HOME = "G28.2"
    PROBE_PLATE = "G38.2"
    GET_PLATE_HEIGHT = "M836"
    GET_CURRENT_POSITION = "M114.2"
    MOVE = "G0"
    DEVICE_INFO = "M115"
    PROGRAMMING_MODE = "dfu"


MAG_DECK_BAUDRATE = 115200

MAG_DECK_COMMAND_TERMINATOR = "\r\n\r\n"
MAG_DECK_ACK = "ok\r\nok\r\n"


# Number of digits after the decimal point for millimeter values
# being sent to/from magnetic module
GCODE_ROUNDING_PRECISION = 3


class MagDeckError(Exception):
    pass


class MagDeckDriver(AbstractMagDeckDriver):
    @classmethod
    async def create(
        cls, port: str, loop: Optional[asyncio.AbstractEventLoop] = None
    ) -> MagDeckDriver:
        """
        Create a mag deck driver.

        Args:
            port: port or url of magdeck
            loop: optional event loop

        Returns: driver
        """
        connection = await SerialConnection.create(
            port=port,
            baud_rate=MAG_DECK_BAUDRATE,
            timeout=DEFAULT_MAG_DECK_TIMEOUT,
            ack=MAG_DECK_ACK,
            loop=loop,
            reset_buffer_before_write=False,
        )
        return cls(connection=connection)

    def __init__(self, connection: SerialConnection):
        """
        Constructor

        Args:
            connection: connection to magdeck.
        """
        self._connection = connection

    async def connect(self) -> None:
        """Connect to device"""
        await self._connection.open()

    async def disconnect(self) -> None:
        """Disconnect from device"""
        await self._connection.close()

    async def is_connected(self) -> bool:
        """Check if connected."""
        return await self._connection.is_open()

    async def home(self) -> None:
        """Homes the magnet"""
        c = CommandBuilder(terminator=MAG_DECK_COMMAND_TERMINATOR).add_gcode(
            gcode=GCODE.HOME
        )
        await self._send_command(c)

    async def probe_plate(self) -> None:
        """
        Probes for the deck plate and calculates the plate distance
        from home.
        To be used for calibrating MagDeck
        """
        c = CommandBuilder(terminator=MAG_DECK_COMMAND_TERMINATOR).add_gcode(
            gcode=GCODE.PROBE_PLATE
        )
        await self._send_command(c)

    async def get_plate_height(self) -> float:
        """
        Default plate_height for the device is 30;
        calculated as MAX_TRAVEL_DISTANCE(45mm) - 15mm
        """
        c = CommandBuilder(terminator=MAG_DECK_COMMAND_TERMINATOR).add_gcode(
            gcode=GCODE.GET_PLATE_HEIGHT
        )
        response = await self._send_command(c)
        data = utils.parse_key_values(response)
        return utils.parse_number(str(data.get("height")), GCODE_ROUNDING_PRECISION)

    async def get_mag_position(self) -> float:
        """
        Default mag_position for the device is 0.0
        i.e. it boots with the current position as 0.0
        """
        c = CommandBuilder(terminator=MAG_DECK_COMMAND_TERMINATOR).add_gcode(
            gcode=GCODE.GET_CURRENT_POSITION
        )

        response = await self._send_command(c)
        data = utils.parse_key_values(response)
        return utils.parse_number(str(data.get("Z")), GCODE_ROUNDING_PRECISION)

    async def move(self, position: float) -> None:
        """
        Move the magnets along Z axis where the home position is 0.0;
        position-> a point along Z. Does not self-check if the position
        is outside of the deck's linear range

        The units of position depend on the module model.
        For GEN1, it's half millimeters ("short millimeters").
        For GEN2, it's millimeters.
        """
        c = (
            CommandBuilder(terminator=MAG_DECK_COMMAND_TERMINATOR)
            .add_gcode(gcode=GCODE.MOVE)
            .add_float(prefix="Z", value=position, precision=GCODE_ROUNDING_PRECISION)
        )
        await self._send_command(c)

    async def get_device_info(self) -> Dict[str, str]:
        """
        Queries Mag-Deck for it's build version, model, and serial number

        Returns: dict
            Where keys are the strings 'version', 'model', and 'serial',
            and each value is a string identifier

            {
                'serial': '1aa11bb22',
                'model': '1aa11bb22',
                'version': '1aa11bb22'
            }

        Example input from Mag-Deck's serial response:
            "serial:aa11bb22 model:aa11bb22 version:aa11bb22"
        """
        c = CommandBuilder(terminator=MAG_DECK_COMMAND_TERMINATOR).add_gcode(
            gcode=GCODE.DEVICE_INFO
        )
        response = await self._send_command(c)
        return utils.parse_device_information(device_info_string=response)

    async def enter_programming_mode(self) -> None:
        """
        Enters and stays in DFU mode for 8 seconds.
        The module resets upon exiting the mode
        which causes the robot to lose serial connection to it.
        The connection can be restored by performing a .disconnect()
        followed by a .connect() to the same symlink node
        """
        c = CommandBuilder(terminator=MAG_DECK_COMMAND_TERMINATOR).add_gcode(
            gcode=GCODE.PROGRAMMING_MODE
        )
        await self._send_command(c)

    async def _send_command(self, command: CommandBuilder) -> str:
        """
        Send the command

        Args:
            command: command to send

        Returns:
            command response
        """
        response = await self._connection.send_command(
            command=command, retries=DEFAULT_COMMAND_RETRIES
        )
        return response



"""Protocol run control and management."""
import asyncio
from typing import List, NamedTuple, Optional, Union

from abc import ABC, abstractmethod

import anyio

from opentrons.hardware_control import HardwareControlAPI
from opentrons import protocol_reader
from opentrons.legacy_broker import LegacyBroker
from opentrons.protocol_reader import (
    ProtocolSource,
    JsonProtocolConfig,
    PythonProtocolConfig,
)
from opentrons.protocol_engine import (
    ProtocolEngine,
    StateSummary,
    Command,
    commands as pe_commands,
)
from opentrons.protocols.parse import PythonParseMode
from opentrons.util.broker import Broker

from .task_queue import TaskQueue
from .json_file_reader import JsonFileReader
from .json_translator import JsonTranslator
from .legacy_context_plugin import LegacyContextPlugin
from .legacy_wrappers import (
    LEGACY_PYTHON_API_VERSION_CUTOFF,
    LEGACY_JSON_SCHEMA_VERSION_CUTOFF,
    LegacyFileReader,
    LegacyContextCreator,
    LegacyExecutor,
    LegacyLoadInfo,
)
from ..protocol_engine.types import PostRunHardwareState, DeckConfigurationType


class RunResult(NamedTuple):
    """Result data from a run, pulled from the ProtocolEngine."""

    commands: List[Command]
    state_summary: StateSummary


class AbstractRunner(ABC):
    """An interface to manage and control a protocol run.

    A Runner is primarily responsible for feeding a ProtocolEngine
    with commands and control signals. These commands and signals are
    generated by protocol files, hardware signals, or externally via
    the HTTP robot-server.

    A Runner controls a single run. Once the run is finished,
    you will need a new Runner to do another run.
    """

    def __init__(self, protocol_engine: ProtocolEngine) -> None:
        self._protocol_engine = protocol_engine
        self._broker = LegacyBroker()

    # TODO(mm, 2023-10-03): `LegacyBroker` is specific to Python protocols and JSON protocols ≤v5.
    # We'll need to extend this in order to report progress from newer JSON protocols.
    #
    # TODO(mm, 2023-10-04): When we switch this to return a new `Broker` instead of a
    # `LegacyBroker`, we should annotate the return type as a `ReadOnlyBroker`.
    @property
    def broker(self) -> LegacyBroker:
        """Return a broker that you can subscribe to in order to monitor protocol progress.

        Currently, this only returns messages for `PythonAndLegacyRunner`.
        Otherwise, it's a no-op.
        """
        return self._broker

    def was_started(self) -> bool:
        """Whether the run has been started.

        This value is latched; once it is True, it will never become False.
        """
        return self._protocol_engine.state_view.commands.has_been_played()

    def play(self, deck_configuration: Optional[DeckConfigurationType] = None) -> None:
        """Start or resume the run."""
        self._protocol_engine.play(deck_configuration=deck_configuration)

    def pause(self) -> None:
        """Pause the run."""
        self._protocol_engine.pause()

    async def stop(self) -> None:
        """Stop (cancel) the run."""
        if self.was_started():
            await self._protocol_engine.stop()
        else:
            await self._protocol_engine.finish(
                drop_tips_after_run=False,
                set_run_status=False,
                post_run_hardware_state=PostRunHardwareState.STAY_ENGAGED_IN_PLACE,
            )

    @abstractmethod
    async def run(
        self,
        deck_configuration: DeckConfigurationType,
        protocol_source: Optional[ProtocolSource] = None,
    ) -> RunResult:
        """Run a given protocol to completion."""


class PythonAndLegacyRunner(AbstractRunner):
    """Protocol runner implementation for Python protocols, and JSON protocols ≤v5."""

    def __init__(
        self,
        protocol_engine: ProtocolEngine,
        hardware_api: HardwareControlAPI,
        task_queue: Optional[TaskQueue] = None,
        legacy_file_reader: Optional[LegacyFileReader] = None,
        legacy_context_creator: Optional[LegacyContextCreator] = None,
        legacy_executor: Optional[LegacyExecutor] = None,
        post_run_hardware_state: PostRunHardwareState = PostRunHardwareState.HOME_AND_STAY_ENGAGED,
        drop_tips_after_run: bool = True,
    ) -> None:
        """Initialize the PythonAndLegacyRunner with its dependencies."""
        super().__init__(protocol_engine)
        self._hardware_api = hardware_api
        self._legacy_file_reader = legacy_file_reader or LegacyFileReader()
        self._legacy_context_creator = legacy_context_creator or LegacyContextCreator(
            hardware_api=hardware_api,
            protocol_engine=protocol_engine,
        )
        self._legacy_executor = legacy_executor or LegacyExecutor()
        # TODO(mc, 2022-01-11): replace task queue with specific implementations
        # of runner interface
        self._task_queue = (
            task_queue or TaskQueue()
        )  # cleanup_func=protocol_engine.finish))
        self._task_queue.set_cleanup_func(
            func=protocol_engine.finish,
            drop_tips_after_run=drop_tips_after_run,
            post_run_hardware_state=post_run_hardware_state,
        )

    async def load(
        self, protocol_source: ProtocolSource, python_parse_mode: PythonParseMode
    ) -> None:
        """Load a Python or JSONv5(& older) ProtocolSource into managed ProtocolEngine."""
        labware_definitions = await protocol_reader.extract_labware_definitions(
            protocol_source=protocol_source
        )
        for definition in labware_definitions:
            # Assume adding a labware definition is fast and there are not many labware
            # definitions, so we don't need to yield here.
            self._protocol_engine.add_labware_definition(definition)

        # fixme(mm, 2022-12-23): This does I/O and compute-bound parsing that will block
        # the event loop. Jira RSS-165.
        protocol = self._legacy_file_reader.read(
            protocol_source, labware_definitions, python_parse_mode
        )
        equipment_broker = None

        if protocol.api_level < LEGACY_PYTHON_API_VERSION_CUTOFF:
            equipment_broker = Broker[LegacyLoadInfo]()
            self._protocol_engine.add_plugin(
                LegacyContextPlugin(
                    broker=self._broker, equipment_broker=equipment_broker
                )
            )
            self._hardware_api.should_taskify_movement_execution(taskify=True)
        else:
            self._hardware_api.should_taskify_movement_execution(taskify=False)

        context = self._legacy_context_creator.create(
            protocol=protocol,
            broker=self._broker,
            equipment_broker=equipment_broker,
        )
        initial_home_command = pe_commands.HomeCreate(
            params=pe_commands.HomeParams(axes=None)
        )
        # this command homes all axes, including pipette plugner and gripper jaw
        self._protocol_engine.add_command(request=initial_home_command)

        self._task_queue.set_run_func(
            func=self._legacy_executor.execute,
            protocol=protocol,
            context=context,
        )

    async def run(  # noqa: D102
        self,
        deck_configuration: DeckConfigurationType,
        protocol_source: Optional[ProtocolSource] = None,
        python_parse_mode: PythonParseMode = PythonParseMode.NORMAL,
    ) -> RunResult:
        # TODO(mc, 2022-01-11): move load to runner creation, remove from `run`
        # currently `protocol_source` arg is only used by tests
        if protocol_source:
            await self.load(
                protocol_source=protocol_source, python_parse_mode=python_parse_mode
            )

        self.play(deck_configuration=deck_configuration)
        self._task_queue.start()
        await self._task_queue.join()

        run_data = self._protocol_engine.state_view.get_summary()
        commands = self._protocol_engine.state_view.commands.get_all()
        return RunResult(commands=commands, state_summary=run_data)


class JsonRunner(AbstractRunner):
    """Protocol runner implementation for json protocols."""

    def __init__(
        self,
        protocol_engine: ProtocolEngine,
        hardware_api: HardwareControlAPI,
        task_queue: Optional[TaskQueue] = None,
        json_file_reader: Optional[JsonFileReader] = None,
        json_translator: Optional[JsonTranslator] = None,
        post_run_hardware_state: PostRunHardwareState = PostRunHardwareState.HOME_AND_STAY_ENGAGED,
        drop_tips_after_run: bool = True,
    ) -> None:
        """Initialize the JsonRunner with its dependencies."""
        super().__init__(protocol_engine)
        self._protocol_engine = protocol_engine
        self._hardware_api = hardware_api
        self._json_file_reader = json_file_reader or JsonFileReader()
        self._json_translator = json_translator or JsonTranslator()
        # TODO(mc, 2022-01-11): replace task queue with specific implementations
        # of runner interface
        self._task_queue = (
            task_queue or TaskQueue()
        )  # cleanup_func=protocol_engine.finish))
        self._task_queue.set_cleanup_func(
            func=protocol_engine.finish,
            drop_tips_after_run=drop_tips_after_run,
            post_run_hardware_state=post_run_hardware_state,
        )

        self._hardware_api.should_taskify_movement_execution(taskify=False)

    async def load(self, protocol_source: ProtocolSource) -> None:
        """Load a JSONv6+ ProtocolSource into managed ProtocolEngine."""
        labware_definitions = await protocol_reader.extract_labware_definitions(
            protocol_source=protocol_source
        )
        for definition in labware_definitions:
            # Assume adding a labware definition is fast and there are not many labware
            # definitions, so we don't need to yield here.
            self._protocol_engine.add_labware_definition(definition)

        protocol = await anyio.to_thread.run_sync(
            self._json_file_reader.read,
            protocol_source,
        )

        commands = await anyio.to_thread.run_sync(
            self._json_translator.translate_commands,
            protocol,
        )

        # Add commands and liquids to the ProtocolEngine.
        #
        # We yield on every iteration so that loading large protocols doesn't block the
        # event loop. With a 24-step 10k-command protocol (See RQA-443), adding all the
        # commands can take 3 to 7 seconds.
        #
        # It wouldn't be safe to do this in a worker thread because each addition
        # invokes the ProtocolEngine's ChangeNotifier machinery, which is not
        # thread-safe.
        liquids = await anyio.to_thread.run_sync(
            self._json_translator.translate_liquids, protocol
        )
        for liquid in liquids:
            self._protocol_engine.add_liquid(
                id=liquid.id,
                name=liquid.displayName,
                description=liquid.description,
                color=liquid.displayColor,
            )
            await _yield()
        initial_home_command = pe_commands.HomeCreate(
            params=pe_commands.HomeParams(axes=None)
        )
        # this command homes all axes, including pipette plugner and gripper jaw
        self._protocol_engine.add_command(request=initial_home_command)

        for command in commands:
            self._protocol_engine.add_command(request=command)
            await _yield()

        self._task_queue.set_run_func(func=self._protocol_engine.wait_until_complete)

    async def run(  # noqa: D102
        self,
        deck_configuration: DeckConfigurationType,
        protocol_source: Optional[ProtocolSource] = None,
    ) -> RunResult:
        # TODO(mc, 2022-01-11): move load to runner creation, remove from `run`
        # currently `protocol_source` arg is only used by tests
        if protocol_source:
            await self.load(protocol_source)

        self.play(deck_configuration=deck_configuration)
        self._task_queue.start()
        await self._task_queue.join()

        run_data = self._protocol_engine.state_view.get_summary()
        commands = self._protocol_engine.state_view.commands.get_all()
        return RunResult(commands=commands, state_summary=run_data)


class LiveRunner(AbstractRunner):
    """Protocol runner implementation for live http protocols."""

    def __init__(
        self,
        protocol_engine: ProtocolEngine,
        hardware_api: HardwareControlAPI,
        task_queue: Optional[TaskQueue] = None,
    ) -> None:
        """Initialize the LiveRunner with its dependencies."""
        super().__init__(protocol_engine)
        self._protocol_engine = protocol_engine
        # TODO(mc, 2022-01-11): replace task queue with specific implementations
        # of runner interface
        self._hardware_api = hardware_api
        self._task_queue = task_queue or TaskQueue()
        self._task_queue.set_cleanup_func(func=protocol_engine.finish)

        self._hardware_api.should_taskify_movement_execution(taskify=False)

    def prepare(self) -> None:
        """Set the task queue to wait until all commands are executed."""
        self._task_queue.set_run_func(func=self._protocol_engine.wait_until_complete)

    async def run(  # noqa: D102
        self,
        deck_configuration: DeckConfigurationType,
        protocol_source: Optional[ProtocolSource] = None,
    ) -> RunResult:
        assert protocol_source is None
        await self._hardware_api.home()
        self.play(deck_configuration=deck_configuration)
        self._task_queue.start()
        await self._task_queue.join()

        run_data = self._protocol_engine.state_view.get_summary()
        commands = self._protocol_engine.state_view.commands.get_all()
        return RunResult(commands=commands, state_summary=run_data)


AnyRunner = Union[PythonAndLegacyRunner, JsonRunner, LiveRunner]


def create_protocol_runner(
    protocol_config: Optional[Union[JsonProtocolConfig, PythonProtocolConfig]],
    protocol_engine: ProtocolEngine,
    hardware_api: HardwareControlAPI,
    task_queue: Optional[TaskQueue] = None,
    json_file_reader: Optional[JsonFileReader] = None,
    json_translator: Optional[JsonTranslator] = None,
    legacy_file_reader: Optional[LegacyFileReader] = None,
    legacy_context_creator: Optional[LegacyContextCreator] = None,
    legacy_executor: Optional[LegacyExecutor] = None,
    post_run_hardware_state: PostRunHardwareState = PostRunHardwareState.HOME_AND_STAY_ENGAGED,
    drop_tips_after_run: bool = True,
) -> AnyRunner:
    """Create a protocol runner."""
    if protocol_config:
        if (
            isinstance(protocol_config, JsonProtocolConfig)
            and protocol_config.schema_version >= LEGACY_JSON_SCHEMA_VERSION_CUTOFF
        ):
            return JsonRunner(
                protocol_engine=protocol_engine,
                hardware_api=hardware_api,
                json_file_reader=json_file_reader,
                json_translator=json_translator,
                task_queue=task_queue,
                post_run_hardware_state=post_run_hardware_state,
                drop_tips_after_run=drop_tips_after_run,
            )
        else:
            return PythonAndLegacyRunner(
                protocol_engine=protocol_engine,
                hardware_api=hardware_api,
                task_queue=task_queue,
                legacy_file_reader=legacy_file_reader,
                legacy_context_creator=legacy_context_creator,
                legacy_executor=legacy_executor,
                post_run_hardware_state=post_run_hardware_state,
                drop_tips_after_run=drop_tips_after_run,
            )

    return LiveRunner(
        protocol_engine=protocol_engine,
        hardware_api=hardware_api,
        task_queue=task_queue,
    )


async def _yield() -> None:
    """Yield execution to the event loop, giving other tasks a chance to run."""
    await asyncio.sleep(0)



"""Pick up tip command request, result, and implementation models."""
from __future__ import annotations
from pydantic import Field
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from ..types import DeckPoint
from .pipetting_common import (
    PipetteIdMixin,
    WellLocationMixin,
    DestinationPositionResult,
)
from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate

if TYPE_CHECKING:
    from ..state import StateView
    from ..execution import MovementHandler, TipHandler


PickUpTipCommandType = Literal["pickUpTip"]


class PickUpTipParams(PipetteIdMixin, WellLocationMixin):
    """Payload needed to move a pipette to a specific well."""

    pass


class PickUpTipResult(DestinationPositionResult):
    """Result data from the execution of a PickUpTip."""

    # Tip volume has a default ONLY for parsing data from earlier versions, which did not include this in the result
    tipVolume: float = Field(
        0,
        description="Maximum volume of liquid that the picked up tip can hold, in µL.",
        ge=0,
    )

    tipLength: float = Field(
        0,
        description="The length of the tip in mm.",
        ge=0,
    )

    tipDiameter: float = Field(
        0,
        description="The diameter of the tip in mm.",
        ge=0,
    )


class PickUpTipImplementation(AbstractCommandImpl[PickUpTipParams, PickUpTipResult]):
    """Pick up tip command implementation."""

    def __init__(
        self,
        state_view: StateView,
        tip_handler: TipHandler,
        movement: MovementHandler,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._tip_handler = tip_handler
        self._movement = movement

    async def execute(self, params: PickUpTipParams) -> PickUpTipResult:
        """Move to and pick up a tip using the requested pipette."""
        pipette_id = params.pipetteId
        labware_id = params.labwareId
        well_name = params.wellName
        well_location = params.wellLocation

        position = await self._movement.move_to_well(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
        )

        tip_geometry = await self._tip_handler.pick_up_tip(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
        )

        return PickUpTipResult(
            tipVolume=tip_geometry.volume,
            tipLength=tip_geometry.length,
            tipDiameter=tip_geometry.diameter,
            position=DeckPoint(x=position.x, y=position.y, z=position.z),
        )


class PickUpTip(BaseCommand[PickUpTipParams, PickUpTipResult]):
    """Pick up tip command model."""

    commandType: PickUpTipCommandType = "pickUpTip"
    params: PickUpTipParams
    result: Optional[PickUpTipResult]

    _ImplementationCls: Type[PickUpTipImplementation] = PickUpTipImplementation


class PickUpTipCreate(BaseCommandCreate[PickUpTipParams]):
    """Pick up tip command creation request model."""

    commandType: PickUpTipCommandType = "pickUpTip"
    params: PickUpTipParams

    _CommandCls: Type[PickUpTip] = PickUpTip



"""The Protocol API v2 context implementation package."""
# TODO(mc, 2022-08.22): uncomment when import cycles can be resolved
# from .legacy_protocol_core import LegacyProtocolCore
# from .legacy_instrument_core import LegacyInstrumentCore
# from .legacy_labware_core.py import LegacyLabwareCore

# __all__ = [
#     "LegacyProtocolCore",
#     "LegacyInstrumentCore",
#     "LegacyLabwareCore",
# ]



from typing import Callable
from typing_extensions import Protocol

from ..types import HardwareEventHandler


class EventSourcer(Protocol):
    """Protocol specifying how to react to events."""

    def register_callback(self, cb: HardwareEventHandler) -> Callable[[], None]:
        """Register a callback that will be called when an event occurs.

        The events may be asynchronous, from various things that can happen
        to the hardware (for instance, the door opening or closing).

        The returned callable removes the callback.
        """
        ...



# noqa: D100

import json
from pathlib import Path
from typing import List

import anyio

from opentrons.protocols.models import LabwareDefinition

from .protocol_source import ProtocolFileRole, ProtocolSource, ProtocolType


async def extract_labware_definitions(
    protocol_source: ProtocolSource,
) -> List[LabwareDefinition]:
    """Extract all the labware definitions that are in the given protocol source.

    This accounts for differences between JSON protocols,
    which embed their labware definitions in the main protocol file,
    and Python protocols, which have them in separate sidecar files.

    May raise an exception if the files are not well-formed. Do not depend on the exact
    exception type. This should not happen if the `ProtocolSource` was acquired from a
    `ProtocolReader`, which should have validated the files.
    """
    if protocol_source.config.protocol_type == ProtocolType.JSON:
        return await _extract_from_json_protocol_file(path=protocol_source.main_file)
        # If there are any separate labware files, ignore them. This avoids custom
        # labware definitions shadowing the ones intrinsic to the main file, which would
        # be confusing. Note that, at least as of v6.2.0, the Opentrons App sends us
        # *every* custom labware definition along with *every* protocol.

    else:
        assert protocol_source.config.protocol_type == ProtocolType.PYTHON
        return [
            await _extract_from_labware_file(file.path)
            for file in protocol_source.files
            if file.role == ProtocolFileRole.LABWARE
        ]


async def _extract_from_labware_file(path: Path) -> LabwareDefinition:
    return await anyio.to_thread.run_sync(LabwareDefinition.parse_file, path)


async def _extract_from_json_protocol_file(path: Path) -> List[LabwareDefinition]:
    def extract_sync(path: Path) -> List[LabwareDefinition]:
        with path.open("rb") as file:
            json_contents = json.load(file)
            # Rely on the file conforming to one of our JSON protocol schemas 3 to 7,
            # which require this labwareDefinitions key.
            unvalidated_definitions = json_contents["labwareDefinitions"].values()
            validated_definitions = [
                LabwareDefinition.parse_obj(u) for u in unvalidated_definitions
            ]
            return validated_definitions

    return await anyio.to_thread.run_sync(extract_sync, path)



from __future__ import annotations
import asyncio
from contextlib import contextmanager, AsyncExitStack
import logging
from typing import (
    Callable,
    Iterator,
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    TYPE_CHECKING,
    Union,
    Sequence,
    cast,
)
from typing_extensions import Final
from pathlib import Path

try:
    import aionotify  # type: ignore[import-untyped]
except (OSError, ModuleNotFoundError):
    aionotify = None

from opentrons_shared_data.pipette import (
    pipette_load_name_conversions as pipette_load_name,
    mutable_configurations,
)
from opentrons_shared_data.pipette.dev_types import PipetteName

from opentrons.drivers.smoothie_drivers import SmoothieDriver
from opentrons.drivers.rpi_drivers import build_gpio_chardev
import opentrons.config
from opentrons.config.types import RobotConfig
from opentrons.types import Mount

from ..module_control import AttachedModulesControl
from ..types import AionotifyEvent, BoardRevision, Axis, DoorState
from ..util import ot2_axis_to_string

if TYPE_CHECKING:
    from opentrons_shared_data.pipette.dev_types import PipetteModel
    from ..dev_types import (
        AttachedPipette,
        AttachedInstruments,
        InstrumentHardwareConfigs,
    )
    from opentrons.drivers.rpi_drivers.dev_types import GPIODriverLike

MODULE_LOG = logging.getLogger(__name__)


class Controller:
    """The concrete instance of the controller for actually controlling
    hardware.
    """

    @classmethod
    async def build(cls, config: RobotConfig) -> Controller:
        """Build a Controller instance.

        Use this factory method rather than the initializer to handle proper
        GPIO initialization.

        :param config: A loaded robot config.
        """
        gpio = build_gpio_chardev("gpiochip0")
        gpio.config_by_board_rev()
        await gpio.setup()
        return cls(config, gpio)

    def __init__(self, config: RobotConfig, gpio: GPIODriverLike):
        """Build a Controller instance.

        Always prefer using :py:meth:`.build` to create an instance of this class. For
        more information on arguments, see that method. If you want to use this
        directly, you must pass in an initialized _and set up_ GPIO driver instance.
        """
        if not opentrons.config.IS_ROBOT:
            MODULE_LOG.warning(
                "This is intended to run on a robot, and while it can connect "
                "to a smoothie via a usb/serial adapter unexpected things "
                "using gpios (such as smoothie reset or light management) "
                "will fail. If you are seeing this message and you are "
                "running on a robot, you need to set the RUNNING_ON_PI "
                "environmental variable to 1."
            )

        self.config = config or opentrons.config.robot_configs.load_ot2()

        self._gpio_chardev: Final = gpio
        self._board_revision: Final = self.gpio_chardev.board_rev
        # We handle our own locks in the hardware controller thank you
        self._smoothie_driver = SmoothieDriver(
            config=self.config, gpio_chardev=self._gpio_chardev
        )
        self._cached_fw_version: Optional[str] = None
        self._module_controls: Optional[AttachedModulesControl] = None
        try:
            self._event_watcher = self._build_event_watcher()
        except AttributeError:
            MODULE_LOG.warning(
                "Failed to initiate aionotify, cannot watch modules "
                "or door, likely because not running on linux"
            )

    @staticmethod
    def _build_event_watcher() -> aionotify.Watcher:
        watcher = aionotify.Watcher()
        watcher.watch(
            alias="modules",
            path="/dev",
            flags=(
                aionotify.Flags.CREATE
                | aionotify.Flags.DELETE
                | aionotify.Flags.MOVED_FROM
                | aionotify.Flags.MOVED_TO
            ),
        )
        return watcher

    @property
    def gpio_chardev(self) -> GPIODriverLike:
        return self._gpio_chardev

    @property
    def board_revision(self) -> BoardRevision:
        return self._board_revision

    @property
    def module_controls(self) -> AttachedModulesControl:
        if not self._module_controls:
            raise AttributeError("Module controls not found.")
        return self._module_controls

    @module_controls.setter
    def module_controls(self, module_controls: AttachedModulesControl) -> None:
        self._module_controls = module_controls

    async def get_serial_number(self) -> Optional[str]:
        try:
            return Path("/var/serial").read_text().strip()
        except OSError:
            return None

    def start_gpio_door_watcher(
        self,
        loop: asyncio.AbstractEventLoop,
        update_door_state: Callable[[DoorState], None],
    ) -> None:
        self.gpio_chardev.start_door_switch_watcher(
            loop=loop, update_door_state=update_door_state
        )

    async def update_position(self) -> Dict[str, float]:
        await self._smoothie_driver.update_position()
        return self._smoothie_driver.position

    def _unhomed_axes(self, axes: Sequence[str]) -> List[str]:
        return list(
            axis
            for axis in axes
            if not self._smoothie_driver.homed_flags.get(axis, False)
        )

    def is_homed(self, axes: Sequence[str]) -> bool:
        return not any(self._unhomed_axes(axes))

    async def move(
        self,
        target_position: Dict[str, float],
        home_flagged_axes: bool = True,
        speed: Optional[float] = None,
        axis_max_speeds: Optional[Dict[str, float]] = None,
    ) -> None:
        async with AsyncExitStack() as cmstack:
            if axis_max_speeds:
                await cmstack.enter_async_context(
                    self._smoothie_driver.restore_axis_max_speed(axis_max_speeds)
                )
            await self._smoothie_driver.move(
                target_position, home_flagged_axes=home_flagged_axes, speed=speed
            )

    async def home(self, axes: Optional[List[str]] = None) -> Dict[str, float]:
        if axes:
            args: Tuple[Any, ...] = ("".join(axes),)
        else:
            args = tuple()
        return await self._smoothie_driver.home(*args)

    async def fast_home(self, axes: Sequence[str], margin: float) -> Dict[str, float]:
        converted_axes = "".join(axes)
        return await self._smoothie_driver.fast_home(converted_axes, margin)

    async def _query_mount(
        self, mount: Mount, expected: Union[PipetteModel, PipetteName, None]
    ) -> AttachedPipette:
        found_model: Optional[
            PipetteModel
        ] = await self._smoothie_driver.read_pipette_model(  # type: ignore
            mount.name.lower()
        )
        if found_model and not pipette_load_name.supported_pipette(found_model):
            # TODO: Consider how to handle this error - it bubbles up now
            # and will cause problems at higher levels
            MODULE_LOG.error(f"Bad model on {mount.name}: {found_model}")
            found_model = None
        found_id = await self._smoothie_driver.read_pipette_id(mount.name.lower())

        if found_model:
            path_to_overrides = opentrons.config.get_opentrons_path(
                "pipette_config_overrides_dir"
            )
            converted_found_model = pipette_load_name.convert_pipette_model(found_model)
            converted_found_name = pipette_load_name.convert_to_pipette_name_type(
                found_model
            )
            config = mutable_configurations.load_with_mutable_configurations(
                converted_found_model, path_to_overrides, found_id
            )
            if expected:
                acceptable = [
                    cast(PipetteName, str(converted_found_name))
                ] + config.pipette_backcompat_names
                if expected not in acceptable:
                    raise RuntimeError(
                        f"mount {mount}: instrument"
                        f" {expected} was requested"
                        f" but {converted_found_model} is present"
                    )
            return {"config": config, "id": found_id}
        else:
            if expected:
                raise RuntimeError(
                    f"mount {mount}: instrument {expected} was"
                    f" requested, but no instrument is present"
                )
            return {"config": None, "id": None}

    async def get_attached_instruments(
        self, expected: Dict[Mount, PipetteName]
    ) -> AttachedInstruments:
        """Find the instruments attached to our mounts.
        :param expected: is ignored, it is just meant to enforce
                          the same interface as the simulator, where
                          required instruments can be manipulated.

        :returns: A dict with mounts as the top-level keys. Each mount value is
            a dict with keys 'model' (containing an instrument model name or
            `None`) and 'id' (containing the serial number of the pipette
            attached to that mount, or `None`). Both mounts will always be
            specified.
        """
        return {
            mount: await self._query_mount(mount, expected.get(mount))
            for mount in Mount.ot2_mounts()
        }

    def set_active_current(self, axis_currents: Dict[Axis, float]) -> None:
        """
        This method sets only the 'active' current, i.e., the current for an
        axis' movement. Smoothie driver automatically resets the current for
        pipette axis to a low current (dwelling current) after each move
        """
        self._smoothie_driver.set_active_current(
            {ot2_axis_to_string(axis): amp for axis, amp in axis_currents.items()}
        )

    @contextmanager
    def save_current(self) -> Iterator[None]:
        self._smoothie_driver.push_active_current()
        try:
            yield
        finally:
            self._smoothie_driver.pop_active_current()

    async def _handle_watch_event(self) -> None:
        try:
            event = await self._event_watcher.get_event()
        except asyncio.IncompleteReadError:
            MODULE_LOG.debug("incomplete read error when quitting watcher")
            return
        if event is not None:
            if "ot_module" in event.name:
                event_name = event.name
                flags = aionotify.Flags.parse(event.flags)
                event_description = AionotifyEvent.build(event_name, flags)
                await self.module_controls.handle_module_appearance(event_description)

    async def watch(self, loop: asyncio.AbstractEventLoop) -> None:
        can_watch = aionotify is not None
        if can_watch:
            await self._event_watcher.setup(loop)

        while can_watch and (not self._event_watcher.closed):
            await self._handle_watch_event()

    async def connect(self, port: Optional[str] = None) -> None:
        """Build driver and connect to it."""
        await self._smoothie_driver.connect(port)
        await self.update_fw_version()

    @property
    def axis_bounds(self) -> Dict[Axis, Tuple[float, float]]:
        """The (minimum, maximum) bounds for each axis."""
        return {
            Axis[ax]: (0, pos)
            for ax, pos in self._smoothie_driver.axis_bounds.items()
            if ax not in "BC"
        }

    @property
    def fw_version(self) -> Optional[str]:
        return self._cached_fw_version

    async def update_fw_version(self) -> None:
        self._cached_fw_version = await self._smoothie_driver.get_fw_version()

    async def update_firmware(
        self, filename: str, loop: asyncio.AbstractEventLoop, modeset: bool
    ) -> str:
        msg = await self._smoothie_driver.update_firmware(filename, loop, modeset)
        await self.update_fw_version()
        return msg

    def engaged_axes(self) -> Dict[str, bool]:
        return self._smoothie_driver.engaged_axes

    async def disengage_axes(self, axes: List[str]) -> None:
        await self._smoothie_driver.disengage_axis("".join(axes))

    def set_lights(self, button: Optional[bool], rails: Optional[bool]) -> None:
        if button is not None:
            self.gpio_chardev.set_button_light(blue=button)
        if rails is not None:
            self.gpio_chardev.set_rail_lights(rails)

    def get_lights(self) -> Dict[str, bool]:
        return {
            "button": self.gpio_chardev.get_button_light()[2],
            "rails": self.gpio_chardev.get_rail_lights(),
        }

    def pause(self) -> None:
        self._smoothie_driver.pause()

    def resume(self) -> None:
        self._smoothie_driver.resume()

    async def halt(self) -> None:
        await self._smoothie_driver.kill()

    async def hard_halt(self) -> None:
        await self._smoothie_driver.hard_halt()

    async def probe(self, axis: str, distance: float) -> Dict[str, float]:
        """Run a probe and return the new position dict"""
        return await self._smoothie_driver.probe_axis(axis, distance)

    async def clean_up(self) -> None:
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            return
        if hasattr(self, "_event_watcher"):
            if loop.is_running() and self._event_watcher:
                self._event_watcher.close()
        if hasattr(self, "gpio_chardev"):
            try:
                if not loop.is_closed():
                    self.gpio_chardev.stop_door_switch_watcher(loop)
            except RuntimeError:
                pass

    async def configure_mount(
        self, mount: Mount, config: InstrumentHardwareConfigs
    ) -> None:
        mount_axis = Axis.by_mount(mount)
        plunger_axis = Axis.of_plunger(mount)

        await self._smoothie_driver.update_steps_per_mm(
            {ot2_axis_to_string(plunger_axis): config["steps_per_mm"]}
        )
        await self._smoothie_driver.update_pipette_config(
            ot2_axis_to_string(mount_axis), {"home": config["home_pos"]}
        )
        await self._smoothie_driver.update_pipette_config(
            ot2_axis_to_string(plunger_axis), {"max_travel": config["max_travel"]}
        )
        self._smoothie_driver.set_dwelling_current(
            {ot2_axis_to_string(plunger_axis): config["idle_current"]}
        )
        ms = config["splits"]
        if ms:
            self._smoothie_driver.configure_splits_for(
                {ot2_axis_to_string(plunger_axis): ms}
            )



from __future__ import annotations

from typing_extensions import Literal, Final, TypedDict
from typing import Optional, List, Sequence, TYPE_CHECKING, Union
from opentrons.hardware_control.modules import ThermocyclerStep

if TYPE_CHECKING:
    from opentrons.protocol_api import InstrumentContext
    from opentrons.protocol_api.labware import Well
    from opentrons.protocol_api._trash_bin import TrashBin
    from opentrons.protocol_api._waste_chute import WasteChute

from opentrons.types import Location


# type for subscriptions
COMMAND: Final = "command"

# Robot #

DELAY: Final = "command.DELAY"
HOME: Final = "command.HOME"
PAUSE: Final = "command.PAUSE"
RESUME: Final = "command.RESUME"
COMMENT: Final = "command.COMMENT"
MOVE_LABWARE: Final = "command.MOVE_LABWARE"

# Pipette #

ASPIRATE: Final = "command.ASPIRATE"
DISPENSE: Final = "command.DISPENSE"
DISPENSE_IN_DISPOSAL_LOCATION: Final = "command.DISPENSE_IN_DISPOSAL_LOCATION"
MIX: Final = "command.MIX"
CONSOLIDATE: Final = "command.CONSOLIDATE"
DISTRIBUTE: Final = "command.DISTRIBUTE"
TRANSFER: Final = "command.TRANSFER"
PICK_UP_TIP: Final = "command.PICK_UP_TIP"
DROP_TIP: Final = "command.DROP_TIP"
DROP_TIP_IN_DISPOSAL_LOCATION: Final = "command.DROP_TIP_IN_DISPOSAL_LOCATION"
BLOW_OUT: Final = "command.BLOW_OUT"
BLOW_OUT_IN_DISPOSAL_LOCATION: Final = "command.BLOW_OUT_IN_DISPOSAL_LOCATION"
AIR_GAP: Final = "command.AIR_GAP"
TOUCH_TIP: Final = "command.TOUCH_TIP"
RETURN_TIP: Final = "command.RETURN_TIP"
MOVE_TO: Final = "command.MOVE_TO"
MOVE_TO_DISPOSAL_LOCATION: Final = "command.MOVE_TO_DISPOSAL_LOCATION"

# Modules #

HEATER_SHAKER_SET_TARGET_TEMPERATURE: Final = (
    "command.HEATER_SHAKER_SET_TARGET_TEMPERATURE"
)
HEATER_SHAKER_WAIT_FOR_TEMPERATURE: Final = "command.HEATER_SHAKER_WAIT_FOR_TEMPERATURE"
HEATER_SHAKER_SET_AND_WAIT_FOR_SHAKE_SPEED: Final = (
    "command.HEATER_SHAKER_SET_AND_WAIT_FOR_SHAKE_SPEED"
)
HEATER_SHAKER_OPEN_LABWARE_LATCH: Final = "command.HEATER_SHAKER_OPEN_LABWARE_LATCH"
HEATER_SHAKER_CLOSE_LABWARE_LATCH: Final = "command.HEATER_SHAKER_CLOSE_LABWARE_LATCH"
HEATER_SHAKER_DEACTIVATE_SHAKER: Final = "command.HEATER_SHAKER_DEACTIVATE_SHAKER"
HEATER_SHAKER_DEACTIVATE_HEATER: Final = "command.HEATER_SHAKER_DEACTIVATE_HEATER"

MAGDECK_CALIBRATE: Final = "command.MAGDECK_CALIBRATE"
MAGDECK_DISENGAGE: Final = "command.MAGDECK_DISENGAGE"
MAGDECK_ENGAGE: Final = "command.MAGDECK_ENGAGE"

TEMPDECK_DEACTIVATE: Final = "command.TEMPDECK_DEACTIVATE"
TEMPDECK_SET_TEMP: Final = "command.TEMPDECK_SET_TEMP"
TEMPDECK_AWAIT_TEMP: Final = "command.TEMPDECK_AWAIT_TEMP"

THERMOCYCLER_OPEN: Final = "command.THERMOCYCLER_OPEN"
THERMOCYCLER_CLOSE: Final = "command.THERMOCYCLER_CLOSE"
THERMOCYCLER_SET_BLOCK_TEMP: Final = "command.THERMOCYCLER_SET_BLOCK_TEMP"
THERMOCYCLER_EXECUTE_PROFILE: Final = "command.THERMOCYCLER_EXECUTE_PROFILE"
THERMOCYCLER_DEACTIVATE: Final = "command.THERMOCYCLER_DEACTIVATE"
THERMOCYCLER_WAIT_FOR_HOLD: Final = "command.THERMOCYCLER_WAIT_FOR_HOLD"
THERMOCYCLER_WAIT_FOR_TEMP: Final = "command.THERMOCYCLER_WAIT_FOR_TEMP"
THERMOCYCLER_WAIT_FOR_LID_TEMP: Final = "command.THERMOCYCLER_WAIT_FOR_LID_TEMP"
THERMOCYCLER_SET_LID_TEMP: Final = "command.THERMOCYCLER_SET_LID_TEMP"
THERMOCYCLER_DEACTIVATE_LID: Final = "command.THERMOCYCLER_DEACTIVATE_LID"
THERMOCYCLER_DEACTIVATE_BLOCK: Final = "command.THERMOCYCLER_DEACTIVATE_BLOCK"


class TextOnlyPayload(TypedDict):
    text: str


class MultiLocationPayload(TypedDict):
    locations: Sequence[Union[Location, Well]]


class OptionalMultiLocationPayload(TypedDict):
    locations: Optional[Sequence[Union[Location, Well]]]


class SingleInstrumentPayload(TypedDict):
    instrument: InstrumentContext


class MultiInstrumentPayload(TypedDict):
    instruments: Sequence[InstrumentContext]


class CommentCommandPayload(TextOnlyPayload):
    pass


class CommentCommand(TypedDict):
    name: Literal["command.COMMENT"]
    payload: CommentCommandPayload


class DelayCommandPayload(TextOnlyPayload):
    minutes: float
    seconds: float


class DelayCommand(TypedDict):
    name: Literal["command.DELAY"]
    payload: DelayCommandPayload


class PauseCommandPayload(TextOnlyPayload):
    userMessage: Optional[str]


class PauseCommand(TypedDict):
    name: Literal["command.PAUSE"]
    payload: PauseCommandPayload


class ResumeCommandPayload(TextOnlyPayload):
    pass


class ResumeCommand(TypedDict):
    name: Literal["command.RESUME"]
    payload: ResumeCommandPayload


class HeaterShakerSetTargetTemperaturePayload(TextOnlyPayload):
    pass


class HeaterShakerSetTargetTemperatureCommand(TypedDict):
    name: Literal["command.HEATER_SHAKER_SET_TARGET_TEMPERATURE"]
    payload: HeaterShakerSetTargetTemperaturePayload


class HeaterShakerWaitForTemperaturePayload(TextOnlyPayload):
    pass


class HeaterShakerWaitForTemperatureCommand(TypedDict):
    name: Literal["command.HEATER_SHAKER_WAIT_FOR_TEMPERATURE"]
    payload: HeaterShakerWaitForTemperaturePayload


class HeaterShakerSetAndWaitForShakeSpeedPayload(TextOnlyPayload):
    pass


class HeaterShakerSetAndWaitForShakeSpeedCommand(TypedDict):
    name: Literal["command.HEATER_SHAKER_SET_AND_WAIT_FOR_SHAKE_SPEED"]
    payload: HeaterShakerSetAndWaitForShakeSpeedPayload


class HeaterShakerOpenLabwareLatchPayload(TextOnlyPayload):
    pass


class HeaterShakerOpenLabwareLatchCommand(TypedDict):
    name: Literal["command.HEATER_SHAKER_OPEN_LABWARE_LATCH"]
    payload: HeaterShakerOpenLabwareLatchPayload


class HeaterShakerCloseLabwareLatchPayload(TextOnlyPayload):
    pass


class HeaterShakerCloseLabwareLatchCommand(TypedDict):
    name: Literal["command.HEATER_SHAKER_CLOSE_LABWARE_LATCH"]
    payload: HeaterShakerCloseLabwareLatchPayload


class HeaterShakerDeactivateShakerPayload(TextOnlyPayload):
    pass


class HeaterShakerDeactivateShakerCommand(TypedDict):
    name: Literal["command.HEATER_SHAKER_DEACTIVATE_SHAKER"]
    payload: HeaterShakerDeactivateShakerPayload


class HeaterShakerDeactivateHeaterPayload(TextOnlyPayload):
    pass


class HeaterShakerDeactivateHeaterCommand(TypedDict):
    name: Literal["command.HEATER_SHAKER_DEACTIVATE_HEATER"]
    payload: HeaterShakerDeactivateHeaterPayload


class MagdeckEngageCommandPayload(TextOnlyPayload):
    pass


class MagdeckEngageCommand(TypedDict):
    name: Literal["command.MAGDECK_ENGAGE"]
    payload: MagdeckEngageCommandPayload


class MagdeckDisengageCommandPayload(TextOnlyPayload):
    pass


class MagdeckDisengageCommand(TypedDict):
    name: Literal["command.MAGDECK_DISENGAGE"]
    payload: MagdeckDisengageCommandPayload


class MagdeckCalibrateCommandPayload(TextOnlyPayload):
    pass


class MagdeckCalibrateCommand(TypedDict):
    name: Literal["command.MAGDECK_CALIBRATE"]
    payload: MagdeckCalibrateCommandPayload


class TempdeckSetTempCommandPayload(TextOnlyPayload):
    celsius: float


class TempdeckSetTempCommand(TypedDict):
    name: Literal["command.TEMPDECK_SET_TEMP"]
    payload: TempdeckSetTempCommandPayload


class TempdeckAwaitTempCommandPayload(TextOnlyPayload):
    celsius: float


class TempdeckAwaitTempCommand(TypedDict):
    name: Literal["command.TEMPDECK_AWAIT_TEMP"]
    payload: TempdeckAwaitTempCommandPayload


class TempdeckDeactivateCommandPayload(TextOnlyPayload):
    pass


class TempdeckDeactivateCommand(TypedDict):
    name: Literal["command.TEMPDECK_DEACTIVATE"]
    payload: TempdeckDeactivateCommandPayload


class ThermocyclerOpenCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerOpenCommand(TypedDict):
    name: Literal["command.THERMOCYCLER_OPEN"]
    payload: ThermocyclerOpenCommandPayload


class ThermocyclerSetBlockTempCommandPayload(TextOnlyPayload):
    temperature: float
    hold_time: Optional[float]


class ThermocyclerSetBlockTempCommand(TypedDict):
    name: Literal["command.THERMOCYCLER_SET_BLOCK_TEMP"]
    payload: ThermocyclerSetBlockTempCommandPayload


class ThermocyclerExecuteProfileCommandPayload(TextOnlyPayload):
    steps: List[ThermocyclerStep]


class ThermocyclerExecuteProfileCommand(TypedDict):
    name: Literal["command.THERMOCYCLER_EXECUTE_PROFILE"]
    payload: ThermocyclerExecuteProfileCommandPayload


class ThermocyclerWaitForHoldCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerWaitForHoldCommand(TypedDict):
    name: Literal["command.THERMOCYCLER_WAIT_FOR_HOLD"]
    payload: ThermocyclerWaitForHoldCommandPayload


class ThermocyclerWaitForTempCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerWaitForTempCommand(TypedDict):
    name: Literal["command.THERMOCYCLER_WAIT_FOR_TEMP"]
    payload: ThermocyclerWaitForTempCommandPayload


class ThermocyclerSetLidTempCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerSetLidTempCommand(TypedDict):
    name: Literal["command.THERMOCYCLER_SET_LID_TEMP"]
    payload: ThermocyclerSetLidTempCommandPayload


class ThermocyclerDeactivateLidCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerDeactivateLidCommand(TypedDict):
    name: Literal["command.THERMOCYCLER_DEACTIVATE_LID"]
    payload: ThermocyclerDeactivateLidCommandPayload


class ThermocyclerDeactivateBlockCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerDeactivateBlockCommand(TypedDict):
    name: Literal["command.THERMOCYCLER_DEACTIVATE_BLOCK"]
    payload: ThermocyclerDeactivateBlockCommandPayload


class ThermocyclerDeactivateCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerDeactivateCommand(TypedDict):
    name: Literal["command.THERMOCYCLER_DEACTIVATE"]
    payload: ThermocyclerDeactivateCommandPayload


class ThermocyclerWaitForLidTempCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerWaitForLidTempCommand(TypedDict):
    name: Literal["command.THERMOCYCLER_WAIT_FOR_LID_TEMP"]
    payload: ThermocyclerWaitForLidTempCommandPayload


class ThermocyclerCloseCommandPayload(TextOnlyPayload):
    pass


class ThermocyclerCloseCommand(TypedDict):
    name: Literal["command.THERMOCYCLER_CLOSE"]
    payload: ThermocyclerCloseCommandPayload


class HomeCommandPayload(TextOnlyPayload):
    axis: str


class HomeCommand(TypedDict):
    name: Literal["command.HOME"]
    payload: HomeCommandPayload


class AspirateDispenseCommandPayload(TextOnlyPayload, SingleInstrumentPayload):
    location: Location
    volume: float
    rate: float


class AspirateCommand(TypedDict):
    name: Literal["command.ASPIRATE"]
    payload: AspirateDispenseCommandPayload


class DispenseCommand(TypedDict):
    name: Literal["command.DISPENSE"]
    payload: AspirateDispenseCommandPayload


class DispenseInDisposalLocationCommandPayload(
    TextOnlyPayload, SingleInstrumentPayload
):
    location: Union[TrashBin, WasteChute]
    volume: float
    rate: float


class DispenseInDisposalLocationCommand(TypedDict):
    name: Literal["command.DISPENSE_IN_DISPOSAL_LOCATION"]
    payload: DispenseInDisposalLocationCommandPayload


class ConsolidateCommandPayload(
    TextOnlyPayload, MultiLocationPayload, SingleInstrumentPayload
):
    volume: Union[float, List[float]]
    source: List[Union[Location, Well]]
    dest: Union[Location, Well]


class ConsolidateCommand(TypedDict):
    name: Literal["command.CONSOLIDATE"]
    payload: ConsolidateCommandPayload


class DistributeCommandPayload(
    TextOnlyPayload, MultiLocationPayload, SingleInstrumentPayload
):
    volume: Union[float, List[float]]
    source: Union[Location, Well]
    dest: List[Union[Location, Well]]


class DistributeCommand(TypedDict):
    name: Literal["command.DISTRIBUTE"]
    payload: DistributeCommandPayload


class TransferCommandPayload(
    TextOnlyPayload, MultiLocationPayload, SingleInstrumentPayload
):
    volume: Union[float, List[float]]
    source: List[Union[Location, Well]]
    dest: List[Union[Location, Well]]


class TransferCommand(TypedDict):
    name: Literal["command.TRANSFER"]
    payload: TransferCommandPayload


class MixCommandPayload(TextOnlyPayload, SingleInstrumentPayload):
    location: Union[None, Location, Well]
    volume: float
    repetitions: int


class MixCommand(TypedDict):
    name: Literal["command.MIX"]
    payload: MixCommandPayload


class BlowOutCommandPayload(TextOnlyPayload, SingleInstrumentPayload):
    location: Optional[Location]


class BlowOutCommand(TypedDict):
    name: Literal["command.BLOW_OUT"]
    payload: BlowOutCommandPayload


class BlowOutInDisposalLocationCommandPayload(TextOnlyPayload, SingleInstrumentPayload):
    location: Union[TrashBin, WasteChute]


class BlowOutInDisposalLocationCommand(TypedDict):
    name: Literal["command.BLOW_OUT_IN_DISPOSAL_LOCATION"]
    payload: BlowOutInDisposalLocationCommandPayload


class TouchTipCommandPayload(TextOnlyPayload, SingleInstrumentPayload):
    pass


class TouchTipCommand(TypedDict):
    name: Literal["command.TOUCH_TIP"]
    payload: TouchTipCommandPayload


class AirGapCommandPayload(TextOnlyPayload):
    pass


class AirGapCommand(TypedDict):
    name: Literal["command.AIR_GAP"]
    payload: AirGapCommandPayload


class ReturnTipCommandPayload(TextOnlyPayload):
    pass


class ReturnTipCommand(TypedDict):
    name: Literal["command.RETURN_TIP"]
    payload: ReturnTipCommandPayload


class PickUpTipCommandPayload(TextOnlyPayload, SingleInstrumentPayload):
    location: Well


class PickUpTipCommand(TypedDict):
    name: Literal["command.PICK_UP_TIP"]
    payload: PickUpTipCommandPayload


class DropTipCommandPayload(TextOnlyPayload, SingleInstrumentPayload):
    location: Well


class DropTipCommand(TypedDict):
    name: Literal["command.DROP_TIP"]
    payload: DropTipCommandPayload


class DropTipInDisposalLocationCommandPayload(TextOnlyPayload, SingleInstrumentPayload):
    location: Union[TrashBin, WasteChute]


class DropTipInDisposalLocationCommand(TypedDict):
    name: Literal["command.DROP_TIP_IN_DISPOSAL_LOCATION"]
    payload: DropTipInDisposalLocationCommandPayload


class MoveToCommandPayload(TextOnlyPayload, SingleInstrumentPayload):
    location: Location


class MoveToCommand(TypedDict):
    name: Literal["command.MOVE_TO"]
    payload: MoveToCommandPayload


class MoveToDisposalLocationCommandPayload(TextOnlyPayload, SingleInstrumentPayload):
    location: Union[TrashBin, WasteChute]


class MoveToDisposalLocationCommand(TypedDict):
    name: Literal["command.MOVE_TO_DISPOSAL_LOCATION"]
    payload: MoveToDisposalLocationCommandPayload


class MoveLabwareCommandPayload(TextOnlyPayload):
    pass


class MoveLabwareCommand(TypedDict):
    name: Literal["command.MOVE_LABWARE"]
    payload: MoveLabwareCommandPayload


Command = Union[
    DropTipCommand,
    DropTipInDisposalLocationCommand,
    PickUpTipCommand,
    ReturnTipCommand,
    AirGapCommand,
    TouchTipCommand,
    BlowOutCommand,
    BlowOutInDisposalLocationCommand,
    MixCommand,
    TransferCommand,
    DistributeCommand,
    ConsolidateCommand,
    DispenseCommand,
    DispenseInDisposalLocationCommand,
    AspirateCommand,
    HomeCommand,
    HeaterShakerSetTargetTemperatureCommand,
    HeaterShakerWaitForTemperatureCommand,
    HeaterShakerSetAndWaitForShakeSpeedCommand,
    HeaterShakerOpenLabwareLatchCommand,
    HeaterShakerCloseLabwareLatchCommand,
    HeaterShakerDeactivateShakerCommand,
    HeaterShakerDeactivateHeaterCommand,
    ThermocyclerCloseCommand,
    ThermocyclerWaitForLidTempCommand,
    ThermocyclerDeactivateCommand,
    ThermocyclerDeactivateBlockCommand,
    ThermocyclerDeactivateLidCommand,
    ThermocyclerSetLidTempCommand,
    ThermocyclerWaitForTempCommand,
    ThermocyclerWaitForHoldCommand,
    ThermocyclerExecuteProfileCommand,
    ThermocyclerSetBlockTempCommand,
    ThermocyclerOpenCommand,
    TempdeckDeactivateCommand,
    TempdeckAwaitTempCommand,
    TempdeckSetTempCommand,
    MagdeckCalibrateCommand,
    MagdeckDisengageCommand,
    MagdeckEngageCommand,
    ResumeCommand,
    PauseCommand,
    DelayCommand,
    CommentCommand,
    MoveToCommand,
    MoveToDisposalLocationCommand,
    MoveLabwareCommand,
]


CommandPayload = Union[
    CommentCommandPayload,
    ResumeCommandPayload,
    HeaterShakerSetTargetTemperaturePayload,
    HeaterShakerWaitForTemperaturePayload,
    HeaterShakerSetAndWaitForShakeSpeedPayload,
    HeaterShakerOpenLabwareLatchPayload,
    HeaterShakerCloseLabwareLatchPayload,
    HeaterShakerDeactivateShakerPayload,
    HeaterShakerDeactivateHeaterPayload,
    MagdeckEngageCommandPayload,
    MagdeckDisengageCommandPayload,
    MagdeckCalibrateCommandPayload,
    ThermocyclerOpenCommandPayload,
    ThermocyclerWaitForHoldCommandPayload,
    ThermocyclerWaitForTempCommandPayload,
    ThermocyclerSetLidTempCommandPayload,
    ThermocyclerDeactivateLidCommandPayload,
    ThermocyclerDeactivateBlockCommandPayload,
    ThermocyclerDeactivateCommandPayload,
    ThermocyclerWaitForLidTempCommand,
    ThermocyclerCloseCommandPayload,
    AirGapCommandPayload,
    ReturnTipCommandPayload,
    DropTipCommandPayload,
    DropTipInDisposalLocationCommandPayload,
    PickUpTipCommandPayload,
    TouchTipCommandPayload,
    BlowOutCommandPayload,
    BlowOutInDisposalLocationCommandPayload,
    MixCommandPayload,
    TransferCommandPayload,
    DistributeCommandPayload,
    ConsolidateCommandPayload,
    AspirateDispenseCommandPayload,
    DispenseInDisposalLocationCommandPayload,
    HomeCommandPayload,
    ThermocyclerExecuteProfileCommandPayload,
    ThermocyclerSetBlockTempCommandPayload,
    TempdeckAwaitTempCommandPayload,
    TempdeckSetTempCommandPayload,
    PauseCommandPayload,
    DelayCommandPayload,
    MoveToCommandPayload,
    MoveToDisposalLocationCommandPayload,
    MoveLabwareCommandPayload,
]


MessageSequenceId = Union[Literal["before"], Literal["after"]]


CommandMessageFields = TypedDict(
    "CommandMessageFields",
    {"$": MessageSequenceId, "id": str, "error": Optional[Exception]},
)


class MoveToMessage(CommandMessageFields, MoveToCommand):
    pass


class MoveToDisposalLocationMessage(
    CommandMessageFields, MoveToDisposalLocationCommand
):
    pass


class DropTipMessage(CommandMessageFields, DropTipCommand):
    pass


class DropTipInDisposalLocationMessage(
    CommandMessageFields, DropTipInDisposalLocationCommand
):
    pass


class PickUpTipMessage(CommandMessageFields, PickUpTipCommand):
    pass


class ReturnTipMessage(CommandMessageFields, ReturnTipCommand):
    pass


class AirGapMessage(CommandMessageFields, AirGapCommand):
    pass


class TouchTipMessage(CommandMessageFields, TouchTipCommand):
    pass


class BlowOutMessage(CommandMessageFields, BlowOutCommand):
    pass


class BlowOutInDisposalLocationMessage(
    CommandMessageFields, BlowOutInDisposalLocationCommand
):
    pass


class MixMessage(CommandMessageFields, MixCommand):
    pass


class TransferMessage(CommandMessageFields, TransferCommand):
    pass


class DistributeMessage(CommandMessageFields, DistributeCommand):
    pass


class ConsolidateMessage(CommandMessageFields, ConsolidateCommand):
    pass


class DispenseMessage(CommandMessageFields, DispenseCommand):
    pass


class DispenseInDisposalLocationMessage(
    CommandMessageFields, DispenseInDisposalLocationCommand
):
    pass


class AspirateMessage(CommandMessageFields, AspirateCommand):
    pass


class HomeMessage(CommandMessageFields, HomeCommand):
    pass


class HeaterShakerSetTargetTemperatureMessage(
    CommandMessageFields, HeaterShakerSetTargetTemperatureCommand
):
    pass


class HeaterShakerWaitForTemperatureMessage(
    CommandMessageFields, HeaterShakerWaitForTemperatureCommand
):
    pass


class HeaterShakerSetAndWaitForShakeSpeedMessage(
    CommandMessageFields, HeaterShakerSetAndWaitForShakeSpeedCommand
):
    pass


class HeaterShakerOpenLabwareLatchMessage(
    CommandMessageFields, HeaterShakerOpenLabwareLatchCommand
):
    pass


class HeaterShakerCloseLabwareLatchMessage(
    CommandMessageFields, HeaterShakerCloseLabwareLatchCommand
):
    pass


class HeaterShakerDeactivateShakerMessage(
    CommandMessageFields, HeaterShakerDeactivateShakerCommand
):
    pass


class HeaterShakerDeactivateHeaterMessage(
    CommandMessageFields, HeaterShakerDeactivateHeaterCommand
):
    pass


class ThermocyclerCloseMessage(CommandMessageFields, ThermocyclerCloseCommand):
    pass


class ThermocyclerWaitForLidTempMessage(
    CommandMessageFields, ThermocyclerWaitForLidTempCommand
):
    pass


class ThermocyclerDeactivateMessage(
    CommandMessageFields, ThermocyclerDeactivateCommand
):
    pass


class ThermocyclerDeactivateBlockMessage(
    CommandMessageFields, ThermocyclerDeactivateBlockCommand
):
    pass


class ThermocyclerDeactivateLidMessage(
    CommandMessageFields, ThermocyclerDeactivateLidCommand
):
    pass


class ThermocyclerSetLidTempMessage(
    CommandMessageFields, ThermocyclerSetLidTempCommand
):
    pass


class ThermocyclerWaitForTempMessage(
    CommandMessageFields, ThermocyclerWaitForTempCommand
):
    pass


class ThermocyclerWaitForHoldMessage(
    CommandMessageFields, ThermocyclerWaitForHoldCommand
):
    pass


class ThermocyclerExecuteProfileMessage(
    CommandMessageFields, ThermocyclerExecuteProfileCommand
):
    pass


class ThermocyclerSetBlockTempMessage(
    CommandMessageFields, ThermocyclerSetBlockTempCommand
):
    pass


class ThermocyclerOpenMessage(CommandMessageFields, ThermocyclerOpenCommand):
    pass


class TempdeckDeactivateMessage(CommandMessageFields, TempdeckDeactivateCommand):
    pass


class TempdeckAwaitTempMessage(CommandMessageFields, TempdeckAwaitTempCommand):
    pass


class TempdeckSetTempMessage(CommandMessageFields, TempdeckSetTempCommand):
    pass


class MagdeckCalibrateMessage(CommandMessageFields, MagdeckCalibrateCommand):
    pass


class MagdeckDisengageMessage(CommandMessageFields, MagdeckDisengageCommand):
    pass


class MagdeckEngageMessage(CommandMessageFields, MagdeckEngageCommand):
    pass


class ResumeMessage(CommandMessageFields, ResumeCommand):
    pass


class PauseMessage(CommandMessageFields, PauseCommand):
    pass


class DelayMessage(CommandMessageFields, DelayCommand):
    pass


class CommentMessage(CommandMessageFields, CommentCommand):
    pass


class MoveLabwareMessage(CommandMessageFields, MoveLabwareCommand):
    pass


CommandMessage = Union[
    DropTipMessage,
    DropTipInDisposalLocationMessage,
    PickUpTipMessage,
    ReturnTipMessage,
    AirGapMessage,
    TouchTipMessage,
    BlowOutMessage,
    BlowOutInDisposalLocationMessage,
    MixMessage,
    TransferMessage,
    DistributeMessage,
    ConsolidateMessage,
    DispenseMessage,
    DispenseInDisposalLocationMessage,
    AspirateMessage,
    HomeMessage,
    HeaterShakerSetTargetTemperatureMessage,
    HeaterShakerWaitForTemperatureMessage,
    HeaterShakerSetAndWaitForShakeSpeedMessage,
    HeaterShakerOpenLabwareLatchMessage,
    HeaterShakerCloseLabwareLatchMessage,
    HeaterShakerDeactivateShakerMessage,
    HeaterShakerDeactivateHeaterMessage,
    ThermocyclerCloseMessage,
    ThermocyclerWaitForLidTempMessage,
    ThermocyclerDeactivateMessage,
    ThermocyclerDeactivateBlockMessage,
    ThermocyclerDeactivateLidMessage,
    ThermocyclerSetLidTempMessage,
    ThermocyclerWaitForTempMessage,
    ThermocyclerWaitForHoldMessage,
    ThermocyclerExecuteProfileMessage,
    ThermocyclerSetBlockTempMessage,
    ThermocyclerOpenMessage,
    TempdeckSetTempMessage,
    TempdeckDeactivateMessage,
    MagdeckEngageMessage,
    MagdeckDisengageMessage,
    MagdeckCalibrateMessage,
    CommentMessage,
    DelayMessage,
    PauseMessage,
    ResumeMessage,
    MoveToMessage,
    MoveToDisposalLocationMessage,
    MoveLabwareMessage,
]



"""Interfaces to provide data and other external system resources.

Classes in this module do not maintain state and can be instantiated
as needed. Some classes may contain solely static methods.
"""
from . import pipette_data_provider
from . import labware_validation
from .model_utils import ModelUtils
from .deck_data_provider import DeckDataProvider, DeckFixedLabware
from .labware_data_provider import LabwareDataProvider
from .module_data_provider import ModuleDataProvider
from .ot3_validation import ensure_ot3_hardware


__all__ = [
    "ModelUtils",
    "LabwareDataProvider",
    "DeckDataProvider",
    "DeckFixedLabware",
    "ModuleDataProvider",
    "ensure_ot3_hardware",
    "pipette_data_provider",
    "labware_validation",
]



import asyncio
from typing import Dict, Optional, Any, List, Union
from typing_extensions import Literal
from dataclasses import dataclass, asdict, field, replace
import json
from pathlib import Path
from warnings import warn

from opentrons.config import robot_configs
from opentrons.config.types import RobotConfig, OT3Config
from opentrons.types import Mount
from opentrons.hardware_control import API, HardwareControlAPI, ThreadManager
from opentrons.hardware_control.types import OT3Mount, HardwareFeatureFlags


# Name and kwargs for a module function
@dataclass(frozen=True)
class ModuleCall:
    function_name: str
    args: List[Any] = field(default_factory=list)
    kwargs: Dict[str, Any] = field(default_factory=dict)


@dataclass(frozen=True)
class OT2SimulatorSetup:
    machine: Literal["OT-2 Standard"] = "OT-2 Standard"
    attached_instruments: Dict[Mount, Dict[str, Optional[str]]] = field(
        default_factory=dict
    )
    attached_modules: Dict[str, List[ModuleCall]] = field(default_factory=dict)
    config: Optional[RobotConfig] = None
    strict_attached_instruments: bool = True


@dataclass(frozen=True)
class OT3SimulatorSetup:
    machine: Literal["OT-3 Standard"] = "OT-3 Standard"
    attached_instruments: Dict[OT3Mount, Dict[str, Optional[str]]] = field(
        default_factory=dict
    )
    attached_modules: Dict[str, List[ModuleCall]] = field(default_factory=dict)
    config: Optional[OT3Config] = None
    strict_attached_instruments: bool = True


SimulatorSetup = Union[OT2SimulatorSetup, OT3SimulatorSetup]


async def _simulator_for_setup(
    setup: SimulatorSetup, loop: Optional[asyncio.AbstractEventLoop]
) -> HardwareControlAPI:
    if setup.machine == "OT-2 Standard":
        return await API.build_hardware_simulator(
            attached_instruments=setup.attached_instruments,
            attached_modules=list(setup.attached_modules.keys()),
            config=setup.config,
            strict_attached_instruments=setup.strict_attached_instruments,
            loop=loop,
            feature_flags=HardwareFeatureFlags.build_from_ff(),
        )
    else:
        from opentrons.hardware_control.ot3api import OT3API

        return await OT3API.build_hardware_simulator(
            attached_instruments=setup.attached_instruments,
            attached_modules=list(setup.attached_modules.keys()),
            config=setup.config,
            strict_attached_instruments=setup.strict_attached_instruments,
            loop=loop,
            feature_flags=HardwareFeatureFlags.build_from_ff(),
        )


async def create_simulator(
    setup: SimulatorSetup, loop: Optional[asyncio.AbstractEventLoop] = None
) -> HardwareControlAPI:
    """Create a simulator"""
    simulator = await _simulator_for_setup(setup, loop)
    for attached_module in simulator.attached_modules:
        calls = setup.attached_modules[attached_module.name()]
        for call in calls:
            f = getattr(attached_module, call.function_name)
            await f(*call.args, **call.kwargs)

    return simulator


async def load_simulator(
    path: Path, loop: Optional[asyncio.AbstractEventLoop] = None
) -> HardwareControlAPI:
    """Create a simulator from a JSON file."""
    return await create_simulator(setup=load_simulator_setup(path), loop=loop)


def _thread_manager_for_setup(
    setup: SimulatorSetup,
) -> ThreadManager[HardwareControlAPI]:
    if setup.machine == "OT-2 Standard":
        return ThreadManager(
            API.build_hardware_simulator,
            attached_instruments=setup.attached_instruments,
            attached_modules=list(setup.attached_modules.keys()),
            config=setup.config,
            strict_attached_instruments=setup.strict_attached_instruments,
            feature_flags=HardwareFeatureFlags.build_from_ff(),
        )
    else:
        from opentrons.hardware_control.ot3api import OT3API

        return ThreadManager(
            OT3API.build_hardware_simulator,
            attached_instruments=setup.attached_instruments,
            attached_modules=list(setup.attached_modules.keys()),
            config=setup.config,
            strict_attached_instruments=setup.strict_attached_instruments,
            feature_flags=HardwareFeatureFlags.build_from_ff(),
        )


async def create_simulator_thread_manager(
    setup: SimulatorSetup,
) -> ThreadManager[HardwareControlAPI]:
    """Create a simulator thread manager from a loaded config."""
    thread_manager = _thread_manager_for_setup(setup)
    await thread_manager.managed_thread_ready_async()

    for attached_module in thread_manager.wrapped().attached_modules:
        calls = setup.attached_modules[attached_module.name()]
        for call in calls:
            f = getattr(attached_module, call.function_name)
            await f(*call.args, **call.kwargs)

    return thread_manager


async def load_simulator_thread_manager(
    path: Path,
) -> ThreadManager[HardwareControlAPI]:
    """Create a simulator wrapped in a ThreadManager from a JSON file."""
    return await create_simulator_thread_manager(load_simulator_setup(path))


def save_simulator_setup(simulator_setup: SimulatorSetup, path: Path) -> None:
    """Write a simulator setup to a file."""
    no_config = replace(simulator_setup, config=None)
    as_dict = asdict(no_config)

    as_dict["config"] = (
        robot_configs.config_to_save(simulator_setup.config)
        if simulator_setup.config
        else None
    )

    if as_dict.get("attached_instruments", None):
        as_dict["attached_instruments"] = {
            mount.name.lower(): data
            for mount, data in as_dict["attached_instruments"].items()
        }

    with path.open("w") as f:
        json.dump(as_dict, f)


def load_simulator_setup(path: Path) -> SimulatorSetup:
    """Load a simulator setup from a file."""
    with path.open() as f:
        obj = json.load(f)

    if "machine" not in obj:
        warn(
            "Simulator configuration does not name a machine, defaulting to OT-2 Standard"
        )
    machine_type = obj.get("machine", "OT-2 Standard")
    if machine_type == "OT-2 Standard":
        return OT2SimulatorSetup(
            **{k: _prepare_for_simulator_setup(k, v) for (k, v) in obj.items()}
        )
    else:
        return OT3SimulatorSetup(
            **{k: _prepare_for_ot3_simulator_setup(k, v) for (k, v) in obj.items()}
        )


def _prepare_for_simulator_setup(key: str, value: Dict[str, Any]) -> Any:
    """Convert value to a SimulatorSetup"""
    if key == "attached_instruments" and value:
        return {Mount[mount.upper()]: data for (mount, data) in value.items()}
    if key == "config" and value:
        return robot_configs.build_config_ot2(value)
    if key == "attached_modules" and value:
        return {k: [ModuleCall(**data) for data in v] for (k, v) in value.items()}
    return value


def _prepare_for_ot3_simulator_setup(key: str, value: Dict[str, Any]) -> Any:
    if key == "attached_instruments" and value:
        return {OT3Mount[mount.upper()]: data for (mount, data) in value.items()}
    if key == "config" and value:
        return robot_configs.build_config_ot3(value)
    if key == "attached_modules" and value:
        return {k: [ModuleCall(**data) for data in v] for (k, v) in value.items()}
    return value



"""Magnetic Module disengage command request, result, and implementation models."""


from __future__ import annotations

from typing import Optional, TYPE_CHECKING
from typing_extensions import Literal, Type

from pydantic import BaseModel, Field

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate

if TYPE_CHECKING:
    from opentrons.protocol_engine.execution import EquipmentHandler
    from opentrons.protocol_engine.state import StateView


DisengageCommandType = Literal["magneticModule/disengage"]


class DisengageParams(BaseModel):
    """Input data to disengage a Magnetic Module's magnets."""

    moduleId: str = Field(
        ...,
        description=(
            "The ID of the Magnetic Module whose magnets you want to disengage,"
            " from a prior `loadModule` command."
        ),
    )


class DisengageResult(BaseModel):
    """The result of a Magnetic Module disengage command."""

    pass


class DisengageImplementation(AbstractCommandImpl[DisengageParams, DisengageResult]):
    """The implementation of a Magnetic Module disengage command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        **unused_dependencies: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment

    async def execute(self, params: DisengageParams) -> DisengageResult:
        """Execute a Magnetic Module disengage command.

        Raises:
            ModuleNotLoadedError: If the given module ID has not been loaded.
            WrongModuleTypeError: If the given module ID has been loaded,
                but it's not a Magnetic Module.
            ModuleNotAttachedError: If the given module ID points to a valid loaded
                Magnetic Module, but that module's hardware wasn't found attached.
        """
        # Allow propagation of ModuleNotLoadedError and WrongModuleTypeError.
        mag_module_substate = self._state_view.modules.get_magnetic_module_substate(
            module_id=params.moduleId
        )
        # Allow propagation of ModuleNotAttachedError.
        hardware_module = self._equipment.get_module_hardware_api(
            mag_module_substate.module_id
        )

        if hardware_module is not None:  # Not virtualizing modules.
            await hardware_module.deactivate()

        return DisengageResult()


class Disengage(BaseCommand[DisengageParams, DisengageResult]):
    """A command to disengage a Magnetic Module's magnets."""

    commandType: DisengageCommandType = "magneticModule/disengage"
    params: DisengageParams
    result: Optional[DisengageResult]

    _ImplementationCls: Type[DisengageImplementation] = DisengageImplementation


class DisengageCreate(BaseCommandCreate[DisengageParams]):
    """A request to create a Magnetic Module disengage command."""

    commandType: DisengageCommandType = "magneticModule/disengage"
    params: DisengageParams

    _CommandCls: Type[Disengage] = Disengage



"""Blow-out command request, result, and implementation models."""
from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from ..types import DeckPoint
from .pipetting_common import (
    PipetteIdMixin,
    FlowRateMixin,
    WellLocationMixin,
    DestinationPositionResult,
)
from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate

from opentrons.hardware_control import HardwareControlAPI


if TYPE_CHECKING:
    from ..execution import MovementHandler, PipettingHandler
    from ..state import StateView

BlowOutCommandType = Literal["blowout"]


class BlowOutParams(PipetteIdMixin, FlowRateMixin, WellLocationMixin):
    """Payload required to blow-out a specific well."""

    pass


class BlowOutResult(DestinationPositionResult):
    """Result data from the execution of a blow-out command."""

    pass


class BlowOutImplementation(AbstractCommandImpl[BlowOutParams, BlowOutResult]):
    """BlowOut command implementation."""

    def __init__(
        self,
        movement: MovementHandler,
        pipetting: PipettingHandler,
        state_view: StateView,
        hardware_api: HardwareControlAPI,
        **kwargs: object,
    ) -> None:
        self._movement = movement
        self._pipetting = pipetting
        self._state_view = state_view
        self._hardware_api = hardware_api

    async def execute(self, params: BlowOutParams) -> BlowOutResult:
        """Move to and blow-out the requested well."""
        x, y, z = await self._movement.move_to_well(
            pipette_id=params.pipetteId,
            labware_id=params.labwareId,
            well_name=params.wellName,
            well_location=params.wellLocation,
        )

        await self._pipetting.blow_out_in_place(
            pipette_id=params.pipetteId, flow_rate=params.flowRate
        )

        return BlowOutResult(position=DeckPoint(x=x, y=y, z=z))


class BlowOut(BaseCommand[BlowOutParams, BlowOutResult]):
    """Blow-out command model."""

    commandType: BlowOutCommandType = "blowout"
    params: BlowOutParams
    result: Optional[BlowOutResult]

    _ImplementationCls: Type[BlowOutImplementation] = BlowOutImplementation


class BlowOutCreate(BaseCommandCreate[BlowOutParams]):
    """Create blow-out command request model."""

    commandType: BlowOutCommandType = "blowout"
    params: BlowOutParams

    _CommandCls: Type[BlowOut] = BlowOut



"""Types that are common across protocols."""

from typing import TypeVar, Union, Type
from opentrons.hardware_control.types import OT3Mount
from opentrons.types import Mount
from opentrons.config.types import RobotConfig, OT3Config


class OT2RobotType:
    pass


class FlexRobotType:
    pass


CalibrationType = TypeVar("CalibrationType")

MountArgType = TypeVar(
    "MountArgType", Mount, Union[OT3Mount, Mount], contravariant=True
)

ConfigType = TypeVar("ConfigType", RobotConfig, OT3Config)

ProtocolRobotType = TypeVar(
    "ProtocolRobotType", Type[FlexRobotType], Type[OT2RobotType], covariant=True
)



"""Command execution module."""

from .command_executor import CommandExecutor
from .create_queue_worker import create_queue_worker
from .equipment import (
    EquipmentHandler,
    LoadedLabwareData,
    LoadedPipetteData,
    LoadedModuleData,
    LoadedConfigureForVolumeData,
)
from .movement import MovementHandler
from .gantry_mover import GantryMover
from .labware_movement import LabwareMovementHandler
from .pipetting import PipettingHandler
from .tip_handler import TipHandler
from .queue_worker import QueueWorker
from .rail_lights import RailLightsHandler
from .run_control import RunControlHandler
from .hardware_stopper import HardwareStopper
from .door_watcher import DoorWatcher
from .status_bar import StatusBarHandler

# .thermocycler_movement_flagger omitted from package's public interface.


__all__ = [
    "CommandExecutor",
    "create_queue_worker",
    "EquipmentHandler",
    "LoadedLabwareData",
    "LoadedPipetteData",
    "LoadedModuleData",
    "LoadedConfigureForVolumeData",
    "MovementHandler",
    "GantryMover",
    "PipettingHandler",
    "TipHandler",
    "LabwareMovementHandler",
    "QueueWorker",
    "RunControlHandler",
    "HardwareStopper",
    "DoorWatcher",
    "RailLightsHandler",
    "StatusBarHandler",
]



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Optional, Union

from opentrons import types
from opentrons.hardware_control.dev_types import PipetteDict
from opentrons.hardware_control.types import HardwareAction
from opentrons.protocols.api_support import instrument as instrument_support
from opentrons.protocols.api_support.labware_like import LabwareLike
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.api_support.definitions import MAX_SUPPORTED_VERSION
from opentrons.protocols.api_support.util import (
    FlowRates,
    PlungerSpeeds,
    APIVersionError,
)
from opentrons.protocols.geometry import planning
from opentrons_shared_data.errors.exceptions import (
    UnexpectedTipRemovalError,
    UnexpectedTipAttachError,
)

from ..._trash_bin import TrashBin
from ..._waste_chute import WasteChute
from opentrons.protocol_api._nozzle_layout import NozzleLayout

from ..instrument import AbstractInstrument

from ..legacy.legacy_well_core import LegacyWellCore
from ..legacy.legacy_module_core import LegacyThermocyclerCore, LegacyHeaterShakerCore

if TYPE_CHECKING:
    from .legacy_protocol_core import LegacyProtocolCoreSimulator


_log = logging.getLogger()

_PRE_2_2_TIP_DROP_HEIGHT_MM = 10
"""In PAPIv2.1 and below, tips are always dropped 10 mm from the bottom of the well."""


class LegacyInstrumentCoreSimulator(AbstractInstrument[LegacyWellCore]):
    """A simulation of an instrument context."""

    def __init__(
        self,
        protocol_interface: LegacyProtocolCoreSimulator,
        pipette_dict: PipetteDict,
        mount: types.Mount,
        instrument_name: str,
        default_speed: float = 400.0,
        api_version: Optional[APIVersion] = None,
    ):
        """Constructor."""
        self._protocol_interface = protocol_interface
        self._mount = mount
        self._pipette_dict = pipette_dict
        self._instrument_name = instrument_name
        self._default_speed = default_speed
        self._api_version = api_version or MAX_SUPPORTED_VERSION

        self._flow_rate = FlowRates(self)
        pipette_state = self.get_hardware_state()
        self._flow_rate.set_defaults(
            aspirate_defaults=pipette_state["default_aspirate_flow_rates"],
            dispense_defaults=pipette_state["default_dispense_flow_rates"],
            blow_out_defaults=pipette_state["default_blow_out_flow_rates"],
            api_level=self._api_version,
        )

        self._plunger_speeds = PlungerSpeeds(self)
        # Cache the maximum instrument height
        self._instrument_max_height = (
            protocol_interface.get_hardware().get_instrument_max_height(self._mount)
        )

    def get_default_speed(self) -> float:
        return self._default_speed

    def set_default_speed(self, speed: float) -> None:
        self._default_speed = speed

    def aspirate(
        self,
        location: types.Location,
        well_core: Optional[LegacyWellCore],
        volume: float,
        rate: float,
        flow_rate: float,
        in_place: bool,
    ) -> None:
        if self.get_current_volume() == 0:
            # Make sure we're at the top of the labware and clear of any
            # liquid to prepare the pipette for aspiration
            if self._api_version < APIVersion(2, 3) or not self.is_ready_to_aspirate():
                if location.labware.is_well:
                    self.move_to(
                        location=location.labware.as_well().top(), well_core=well_core
                    )
                else:
                    # TODO(seth,2019/7/29): This should be a warning exposed
                    #  via rpc to the runapp
                    _log.warning(
                        "When aspirate is called on something other than a "
                        "well relative position, we can't move to the top of"
                        " the well to prepare for aspiration. This might "
                        "cause over aspiration if the previous command is a "
                        "blow_out."
                    )
                self.prepare_to_aspirate()
            self.move_to(location=location, well_core=well_core)
        elif location != self._protocol_interface.get_last_location():
            self.move_to(location=location, well_core=well_core)

        self._raise_if_no_tip(HardwareAction.ASPIRATE.name)
        new_volume = self.get_current_volume() + volume
        assert (
            new_volume <= self._pipette_dict["working_volume"]
        ), "Cannot aspirate more than pipette max volume"
        self._pipette_dict["ready_to_aspirate"] = True
        self._update_volume(new_volume)

    def dispense(
        self,
        location: Union[types.Location, TrashBin, WasteChute],
        well_core: Optional[LegacyWellCore],
        volume: float,
        rate: float,
        flow_rate: float,
        in_place: bool,
        push_out: Optional[float],
    ) -> None:
        if isinstance(location, (TrashBin, WasteChute)):
            raise APIVersionError(
                "Dispense in Moveable Trash or Waste Chute are not supported in this API Version."
            )
        if not in_place:
            self.move_to(location=location, well_core=well_core)
        self._raise_if_no_tip(HardwareAction.DISPENSE.name)
        self._update_volume(self.get_current_volume() - volume)

    def blow_out(
        self,
        location: Union[types.Location, TrashBin, WasteChute],
        well_core: Optional[LegacyWellCore],
        in_place: bool,
    ) -> None:
        if isinstance(location, (TrashBin, WasteChute)):
            raise APIVersionError(
                "Blow Out in Moveable Trash or Waste Chute are not supported in this API Version."
            )
        if not in_place:
            self.move_to(location=location, well_core=well_core)
        self._raise_if_no_tip(HardwareAction.BLOWOUT.name)
        self._update_volume(0)
        self._pipette_dict["ready_to_aspirate"] = False

    def _update_volume(self, vol: float) -> None:
        self._pipette_dict["current_volume"] = vol
        self._pipette_dict["available_volume"] = (
            self._pipette_dict["working_volume"] - vol
        )

    def touch_tip(
        self,
        location: types.Location,
        well_core: LegacyWellCore,
        radius: float,
        z_offset: float,
        speed: float,
    ) -> None:
        self.move_to(location)

    def pick_up_tip(
        self,
        location: types.Location,
        well_core: LegacyWellCore,
        presses: Optional[int],
        increment: Optional[float],
        prep_after: bool = True,
    ) -> None:
        geometry = well_core.geometry
        tip_rack_core = geometry.parent
        tip_length = instrument_support.tip_length_for(
            pipette=self._pipette_dict,
            tip_rack_definition=tip_rack_core.get_definition(),
        )

        self.move_to(location=location)
        self._raise_if_tip("pick up tip")
        self._pipette_dict["has_tip"] = True
        self._pipette_dict["tip_length"] = tip_length
        self._pipette_dict["current_volume"] = 0
        self._pipette_dict["working_volume"] = min(
            geometry.max_volume, self.get_max_volume()
        )
        self._pipette_dict["available_volume"] = self._pipette_dict["working_volume"]
        if prep_after:
            self._pipette_dict["ready_to_aspirate"] = True

        tip_rack_core.get_tip_tracker().use_tips(
            start_well=well_core,
            num_channels=self.get_channels(),
            fail_if_full=self._api_version < APIVersion(2, 2),
        )

    def drop_tip(
        self,
        location: Optional[types.Location],
        well_core: LegacyWellCore,
        home_after: Optional[bool],
        alternate_drop_location: Optional[bool] = False,
    ) -> None:
        if alternate_drop_location:
            raise APIVersionError(
                "Tip drop alternation is not supported in this API version."
            )
        labware_core = well_core.geometry.parent

        if location is None:
            from opentrons.protocol_api.labware import Labware, Well

            labware = Labware(
                core=labware_core,
                api_version=self._api_version,
                protocol_core=None,  # type: ignore[arg-type]
                core_map=None,  # type: ignore[arg-type]
            )
            well = Well(parent=labware, core=well_core, api_version=self._api_version)

            if LabwareLike(labware).is_fixed_trash():
                location = well.top()
            elif self._api_version < APIVersion(2, 2):
                location = well.bottom(z=_PRE_2_2_TIP_DROP_HEIGHT_MM)
            else:
                assert (
                    labware_core.is_tip_rack()
                ), "Expected tip drop target to be a tip rack."

                return_height = self.get_return_height()
                location = well.top(z=-return_height * labware_core.get_tip_length())

        self.move_to(location=location)
        self._raise_if_no_tip(HardwareAction.DROPTIP.name)
        self._pipette_dict["has_tip"] = False
        self._pipette_dict["tip_length"] = 0.0
        self._update_volume(0)

        if self._api_version < APIVersion(2, 2) and labware_core.is_tip_rack():
            # If this is a tiprack we can try and add the dirty tip back to the tracker
            try:
                labware_core.get_tip_tracker().return_tips(
                    start_well=well_core,
                    num_channels=self.get_channels(),
                )
            except AssertionError:
                # Similarly to :py:meth:`return_tips`, the failure case here
                # just means the tip can't be reused, so don't actually stop
                # the protocol
                _log.warning(
                    f"Could not return tip to {labware_core.get_display_name()}"
                )

    def drop_tip_in_disposal_location(
        self, disposal_location: Union[TrashBin, WasteChute], home_after: Optional[bool]
    ) -> None:
        raise APIVersionError(
            "Dropping tips in a trash bin or waste chute is not supported in this API Version."
        )

    def home(self) -> None:
        self._protocol_interface.set_last_location(None)

    def home_plunger(self) -> None:
        pass

    def move_to(
        self,
        location: Union[types.Location, TrashBin, WasteChute],
        well_core: Optional[LegacyWellCore] = None,
        force_direct: bool = False,
        minimum_z_height: Optional[float] = None,
        speed: Optional[float] = None,
    ) -> None:
        """Simulation of only the motion planning portion of move_to."""
        if isinstance(location, (TrashBin, WasteChute)):
            raise APIVersionError(
                "Move To Trash Bin and Waste Chute are not supported in this API Version."
            )

        self.flag_unsafe_move(location)

        last_location = self._protocol_interface.get_last_location()
        if last_location:
            from_loc = last_location
        else:
            from_loc = types.Location(types.Point(0, 0, 0), LabwareLike(None))

        # We just want to catch planning errors.
        planning.plan_moves(
            from_loc,
            location,
            self._protocol_interface.get_deck(),
            self._instrument_max_height,
            force_direct=force_direct,
            minimum_z_height=minimum_z_height,
        )

        self._protocol_interface.set_last_location(location)

    def get_mount(self) -> types.Mount:
        return self._mount

    def get_requested_as_name(self) -> str:
        return self._instrument_name

    def get_pipette_name(self) -> str:
        return self._pipette_dict["name"]

    def get_model(self) -> str:
        return self._pipette_dict["model"]

    def get_display_name(self) -> str:
        return self._pipette_dict["display_name"]

    def get_min_volume(self) -> float:
        return self._pipette_dict["min_volume"]

    def get_max_volume(self) -> float:
        return self._pipette_dict["max_volume"]

    def get_working_volume(self) -> float:
        return self._pipette_dict["working_volume"]

    def get_current_volume(self) -> float:
        return self._pipette_dict["current_volume"]

    def get_available_volume(self) -> float:
        return self._pipette_dict["available_volume"]

    def get_hardware_state(self) -> PipetteDict:
        return self._pipette_dict

    def get_channels(self) -> int:
        return self._pipette_dict["channels"]

    def has_tip(self) -> bool:
        return self._pipette_dict["has_tip"]

    def is_ready_to_aspirate(self) -> bool:
        return self._pipette_dict["ready_to_aspirate"]

    def prepare_to_aspirate(self) -> None:
        self._raise_if_no_tip(HardwareAction.PREPARE_ASPIRATE.name)

    def get_return_height(self) -> float:
        return self._pipette_dict["return_tip_height"]

    def get_speed(self) -> PlungerSpeeds:
        return self._plunger_speeds

    def get_flow_rate(self) -> FlowRates:
        return self._flow_rate

    def get_aspirate_flow_rate(self, rate: float = 1.0) -> float:
        return self._pipette_dict["aspirate_flow_rate"] * rate

    def get_dispense_flow_rate(self, rate: float = 1.0) -> float:
        return self._pipette_dict["dispense_flow_rate"] * rate

    def get_blow_out_flow_rate(self, rate: float = 1.0) -> float:
        return self._pipette_dict["blow_out_flow_rate"] * rate

    def set_flow_rate(
        self,
        aspirate: Optional[float] = None,
        dispense: Optional[float] = None,
        blow_out: Optional[float] = None,
    ) -> None:
        self._protocol_interface.get_hardware().set_flow_rate(
            mount=self._mount,
            aspirate=aspirate,
            dispense=dispense,
            blow_out=blow_out,
        )
        self._update_flow_rate()

    def set_pipette_speed(
        self,
        aspirate: Optional[float] = None,
        dispense: Optional[float] = None,
        blow_out: Optional[float] = None,
    ) -> None:
        self._protocol_interface.get_hardware().set_pipette_speed(
            mount=self._mount,
            aspirate=aspirate,
            dispense=dispense,
            blow_out=blow_out,
        )
        self._update_flow_rate()

    def flag_unsafe_move(self, location: types.Location) -> None:
        """Check if a movement to a destination is potentially unsafe.

        Args:
            location: The movement destination.

        Raises:
            RuntimeError: The movement is unsafe.
        """
        from_loc = self._protocol_interface.get_last_location()

        if not from_loc:
            from_loc = types.Location(types.Point(0, 0, 0), LabwareLike(None))

        for mod in self._protocol_interface.get_module_cores():
            if isinstance(mod, LegacyThermocyclerCore):
                mod.flag_unsafe_move(to_loc=location, from_loc=from_loc)
            elif isinstance(mod, LegacyHeaterShakerCore):
                mod.flag_unsafe_move(
                    to_loc=location,
                    is_multichannel=self.get_channels() > 1,
                )

    def _update_flow_rate(self) -> None:
        """Update cached speed and flow rates from hardware controller pipette."""
        p = self._protocol_interface.get_hardware().get_attached_instrument(
            self.get_mount()
        )
        self._pipette_dict["aspirate_flow_rate"] = p["aspirate_flow_rate"]
        self._pipette_dict["dispense_flow_rate"] = p["dispense_flow_rate"]
        self._pipette_dict["blow_out_flow_rate"] = p["blow_out_flow_rate"]
        self._pipette_dict["aspirate_speed"] = p["aspirate_speed"]
        self._pipette_dict["dispense_speed"] = p["dispense_speed"]
        self._pipette_dict["blow_out_speed"] = p["blow_out_speed"]

    def _raise_if_no_tip(self, action: str) -> None:
        """Raise UnexpectedTipRemovalError if no tip."""
        if not self.has_tip():
            raise UnexpectedTipRemovalError(
                action, self._instrument_name, self._mount.name
            )

    def _raise_if_tip(self, action: str) -> None:
        """Raise UnexpectedTipAttachError if tip."""
        if self.has_tip():
            raise UnexpectedTipAttachError(
                action, self._instrument_name, self._mount.name
            )

    def configure_for_volume(self, volume: float) -> None:
        """This will never be called because it was added in API 2.15."""
        pass

    def configure_nozzle_layout(
        self,
        style: NozzleLayout,
        primary_nozzle: Optional[str],
        front_right_nozzle: Optional[str],
    ) -> None:
        """This will never be called because it was added in API 2.15."""
        pass

    def get_active_channels(self) -> int:
        """This will never be called because it was added in API 2.16."""
        assert False, "get_active_channels only supported in API 2.16 & later"

    def is_tip_tracking_available(self) -> bool:
        # Tip tracking is always available in legacy context
        return True



import typing

from dataclasses import dataclass
from datetime import datetime
from enum import Enum

AttitudeMatrix = typing.List[typing.List[float]]


class SourceType(str, Enum):
    """Calibration source type"""

    default = "default"
    factory = "factory"
    user = "user"
    calibration_check = "calibration_check"
    legacy = "legacy"
    unknown = "unknown"


class TipLengthCalNotFound(Exception):
    pass


@dataclass
class CalibrationStatus:
    markedBad: bool = False
    source: typing.Optional[SourceType] = None
    markedAt: typing.Optional[datetime] = None


@dataclass
class UriDetails:
    namespace: str
    load_name: str
    version: int


# TODO(mm, 2023-11-20): Deduplicate this with similar types in robot_server
# and opentrons.protocol_engine.
@dataclass
class CutoutFixturePlacement:
    cutout_fixture_id: str
    cutout_id: str



"""Main ProtocolEngine factory."""
import asyncio
import contextlib
import typing

from opentrons.hardware_control import HardwareControlAPI
from opentrons.hardware_control.types import DoorState
from opentrons.util.async_helpers import async_context_manager_in_thread

from .protocol_engine import ProtocolEngine
from .resources import DeckDataProvider, ModuleDataProvider
from .state import Config, StateStore
from .types import PostRunHardwareState, DeckConfigurationType


# TODO(mm, 2023-06-16): Arguably, this not being a context manager makes us prone to forgetting to
# clean it up properly, especially in tests. See e.g. https://opentrons.atlassian.net/browse/RSS-222
async def create_protocol_engine(
    hardware_api: HardwareControlAPI,
    config: Config,
    load_fixed_trash: bool = False,
    deck_configuration: typing.Optional[DeckConfigurationType] = None,
) -> ProtocolEngine:
    """Create a ProtocolEngine instance.

    Arguments:
        hardware_api: Hardware control API to pass down to dependencies.
        config: ProtocolEngine configuration.
        load_fixed_trash: Automatically load fixed trash labware in engine.
        deck_configuration: The initial deck configuration the engine will be instantiated with.
    """
    deck_data = DeckDataProvider(config.deck_type)
    deck_definition = await deck_data.get_deck_definition()
    deck_fixed_labware = (
        await deck_data.get_deck_fixed_labware(deck_definition)
        if load_fixed_trash
        else []
    )
    module_calibration_offsets = ModuleDataProvider.load_module_calibrations()

    state_store = StateStore(
        config=config,
        deck_definition=deck_definition,
        deck_fixed_labware=deck_fixed_labware,
        is_door_open=hardware_api.door_state is DoorState.OPEN,
        module_calibration_offsets=module_calibration_offsets,
        deck_configuration=deck_configuration,
    )

    return ProtocolEngine(state_store=state_store, hardware_api=hardware_api)


@contextlib.contextmanager
def create_protocol_engine_in_thread(
    hardware_api: HardwareControlAPI,
    config: Config,
    drop_tips_after_run: bool,
    post_run_hardware_state: PostRunHardwareState,
    load_fixed_trash: bool = False,
) -> typing.Generator[
    typing.Tuple[ProtocolEngine, asyncio.AbstractEventLoop], None, None
]:
    """Run a `ProtocolEngine` in a worker thread.

    When this context manager is entered, it:

    1. Starts a worker thread.
    2. Starts an asyncio event loop in that worker thread.
    3. Creates and `.play()`s a `ProtocolEngine` in that event loop.
    4. Returns the `ProtocolEngine` and the event loop.
       Use functions like `asyncio.run_coroutine_threadsafe()` to safely interact with
       the `ProtocolEngine` from your thread.

    When this context manager is exited, it:

    1. Cleans up the `ProtocolEngine`.
    2. Stops and cleans up the event loop.
    3. Joins the thread.
    """
    with async_context_manager_in_thread(
        _protocol_engine(
            hardware_api,
            config,
            drop_tips_after_run,
            post_run_hardware_state,
            load_fixed_trash,
        )
    ) as (
        protocol_engine,
        loop,
    ):
        yield protocol_engine, loop


@contextlib.asynccontextmanager
async def _protocol_engine(
    hardware_api: HardwareControlAPI,
    config: Config,
    drop_tips_after_run: bool,
    post_run_hardware_state: PostRunHardwareState,
    load_fixed_trash: bool = False,
) -> typing.AsyncGenerator[ProtocolEngine, None]:
    protocol_engine = await create_protocol_engine(
        hardware_api=hardware_api,
        config=config,
        load_fixed_trash=load_fixed_trash,
    )
    try:
        # TODO(mm, 2023-11-21): Callers like opentrons.execute need to be able to pass in
        # the deck_configuration argument to ProtocolEngine.play().
        # https://opentrons.atlassian.net/browse/RSS-400
        protocol_engine.play()
        yield protocol_engine
    finally:
        await protocol_engine.finish(
            drop_tips_after_run=drop_tips_after_run,
            post_run_hardware_state=post_run_hardware_state,
        )



from typing import Optional


class Simulation:
    def tick(self) -> None:
        pass


class Temperature(Simulation):
    """A model with a current and target temperature. The current temperature is
    always moving towards the target.
    """

    def __init__(self, per_tick: float, current: float) -> None:
        """Construct a temperature simulation.

        Args:
            per_tick: amount to move per tick,
            current: the starting temperature
        """
        self._per_tick = per_tick
        self._current = current
        self._target: Optional[float] = None

    def tick(self) -> None:
        if self._target is None:
            return

        diff = self._target - self._current

        if abs(diff) < self._per_tick:
            self._current = self._target
        elif diff > 0:
            self._current += self._per_tick
        else:
            self._current -= self._per_tick

    def deactivate(self, temperature: float) -> None:
        """Deactivate and reset to temperature"""
        self._target = None
        self._current = temperature

    def set_target(self, target: Optional[float]) -> None:
        self._target = target

    @property
    def current(self) -> float:
        return self._current

    @property
    def target(self) -> Optional[float]:
        return self._target


class RPM(Simulation):
    """A model with a current and target rpm. The current rpm is
    always moving towards the target.
    """

    def __init__(self, per_tick: float, current: float) -> None:
        """Construct a rpm simulation.

        Args:
            per_tick: amount to move per tick,
            current: the starting rpm
        """
        self._per_tick = per_tick
        self._current = current
        self._target: Optional[float] = None

    def tick(self) -> None:

        if self._target is None:
            target = 0.0
        else:
            target = self._target

        diff = target - self._current

        if abs(diff) < self._per_tick:
            self._current = target
        elif diff > 0:
            self._current += self._per_tick
        else:
            self._current -= self._per_tick

    def deactivate(self, rpm: float) -> None:
        """Deactivate and reset to rpm"""
        self._target = None
        self._current = rpm

    def set_target(self, target: Optional[float]) -> None:
        self._target = target

    @property
    def current(self) -> float:
        return self._current

    @property
    def target(self) -> Optional[float]:
        return self._target


class TemperatureWithHold(Temperature):
    """A model with a current temperature, target temperature, and hold time.
    The current temperature is always moving towards the target.

    When the current temperature is within close enough from target the hold time
    decrements once per tick.
    """

    def __init__(self, per_tick: float, current: float) -> None:
        """Construct a temperature with hold simulation."""
        super().__init__(per_tick=per_tick, current=current)
        self._total_hold: Optional[float] = None
        self._hold: Optional[float] = None

    def tick(self) -> None:
        super().tick()
        if self.target == self._current and self._hold is not None:
            self._hold = max(0, self._hold - 1)

    def set_hold(self, hold: float) -> None:
        self._total_hold = hold
        self._hold = hold

    @property
    def time_remaining(self) -> Optional[float]:
        return self._hold

    @property
    def total_hold(self) -> Optional[float]:
        return self._total_hold



from dataclasses import dataclass
from typing import Optional


@dataclass(frozen=True)
class Liquid:
    """A liquid to load into a well.

    Attributes:
        name: A human-readable name for the liquid.
        description: An optional description.
        display_color: An optional display color for the liquid.

    .. versionadded:: 2.14
    """

    _id: str
    name: str
    description: Optional[str]
    display_color: Optional[str]



import typing
from typing_extensions import Literal
from dataclasses import dataclass
from datetime import datetime
from opentrons.config.robot_configs import default_pipette_offset

from opentrons.calibration_storage import ot2 as calibration_storage
from opentrons.calibration_storage import types, helpers
from opentrons.types import Mount, Point
from opentrons.hardware_control.types import OT3Mount

from opentrons_shared_data.labware.labware_definition import LabwareDefinition

if typing.TYPE_CHECKING:
    from opentrons_shared_data.pipette.dev_types import LabwareUri
    from opentrons_shared_data.labware.dev_types import (
        LabwareDefinition as TypeDictLabwareDef,
    )

# These type aliases aid typechecking in tests that work the same on this and
# the hardware_control.instruments.ot3 variant
SourceType = types.SourceType
CalibrationStatus = types.CalibrationStatus


@dataclass
class PipetteOffsetByPipetteMount:
    """
    Class to store pipette offset without pipette and mount info
    """

    offset: Point
    source: SourceType
    status: CalibrationStatus
    tiprack: typing.Optional[str] = None
    uri: typing.Optional[str] = None
    last_modified: typing.Optional[datetime] = None


@dataclass
class PipetteOffsetCalibration:
    """
    Class to store pipette offset calibration with pipette and mount info
    """

    pipette: str
    mount: str
    offset: Point
    tiprack: str
    uri: str
    last_modified: datetime
    source: SourceType
    status: CalibrationStatus


@dataclass
class TipLengthCalibration:
    tip_length: float
    source: SourceType
    status: CalibrationStatus
    pipette: str
    tiprack: str
    last_modified: datetime
    uri: typing.Union["LabwareUri", Literal[""]]


def load_pipette_offset(
    pip_id: typing.Optional[str], mount: typing.Union[Mount, OT3Mount]
) -> PipetteOffsetByPipetteMount:
    # load default if pipette offset data do not exist
    pip_cal_obj = PipetteOffsetByPipetteMount(
        offset=Point(*default_pipette_offset()),
        source=SourceType.default,
        status=CalibrationStatus(),
    )
    # TODO this can be removed once we switch to using
    # ot3 pipette types in the ot3 hardware controller.
    if isinstance(mount, OT3Mount):
        checked_mount = mount.to_mount()
    else:
        checked_mount = mount
    if pip_id:
        pip_offset_data = calibration_storage.get_pipette_offset(pip_id, checked_mount)
        if pip_offset_data:
            return PipetteOffsetByPipetteMount(
                offset=pip_offset_data.offset,
                tiprack=pip_offset_data.tiprack,
                last_modified=pip_offset_data.last_modified,
                uri=pip_offset_data.uri,
                source=pip_offset_data.source,
                status=types.CalibrationStatus(
                    markedAt=pip_offset_data.status.markedAt,
                    markedBad=pip_offset_data.status.markedBad,
                    source=pip_offset_data.status.source,
                ),
            )
    return pip_cal_obj


def save_pipette_offset_calibration(
    pip_id: typing.Optional[str],
    mount: typing.Union[Mount, OT3Mount],
    offset: Point,
    tiprack_hash: str,
    tiprack_uri: str,
) -> None:
    # TODO this can be removed once we switch to using
    # ot3 pipette types in the ot3 hardware controller.
    if isinstance(mount, OT3Mount):
        checked_mount = mount.to_mount()
    else:
        checked_mount = mount
    if pip_id:
        calibration_storage.save_pipette_calibration(
            offset, pip_id, checked_mount, tiprack_hash, tiprack_uri
        )


# TODO (lc 09-26-2022) We should ensure that only LabwareDefinition models are passed
# into this function instead of a mixture of TypeDicts and BaseModels
def load_tip_length_for_pipette(
    pipette_id: str, tiprack: typing.Union["TypeDictLabwareDef", LabwareDefinition]
) -> TipLengthCalibration:
    if isinstance(tiprack, LabwareDefinition):
        tiprack = typing.cast(
            "TypeDictLabwareDef", tiprack.dict(exclude_none=True, exclude_unset=True)
        )

    tip_length_data = calibration_storage.load_tip_length_calibration(
        pipette_id, tiprack
    )

    tiprack_uri = helpers.uri_from_definition(tiprack)

    return TipLengthCalibration(
        tip_length=tip_length_data.tipLength,
        source=tip_length_data.source,
        pipette=pipette_id,
        tiprack=tip_length_data.definitionHash,
        last_modified=tip_length_data.lastModified,
        uri=tiprack_uri,
        status=types.CalibrationStatus(
            markedAt=tip_length_data.status.markedAt,
            markedBad=tip_length_data.status.markedBad,
            source=tip_length_data.status.source,
        ),
    )
